__all__ = [
    "BaseError",
    "BadRequestError",
    "ConflictError",
    "ForbiddenError",
    "InternalError",
    "LoadError",
    "NotFoundError",
    "NotModified",
    "NotSupportedError",
    "PreconditionFailedError",
    "SpecError",
    "UnauthorizedError",
]


class BaseError(Exception):
    status_code: int


class NotModified(BaseError):
    status_code = 304


class BadRequestError(BaseError):
    status_code = 400


class UnauthorizedError(BaseError):
    status_code = 401


class ForbiddenError(BaseError):
    status_code = 403


class NotFoundError(BaseError):
    status_code = 404


class ConflictError(BaseError):
    status_code = 409


class PreconditionFailedError(BaseError):
    status_code = 412


class NotSupportedError(BaseError):
    status_code = 415


class InternalError(Exception):
    status_code = 500


class LoadError(Exception):
    status_code = 500


class SpecError(Exception):
    status_code = 500



import asyncio
import threading
from typing import Any, Callable

_loop = None
_loop_thread = None


def _ensure_loop():
    global _loop, _loop_thread
    if _loop and not _loop.is_closed():
        return _loop

    _loop = asyncio.new_event_loop()

    def _run_loop():
        asyncio.set_event_loop(_loop)
        _loop.run_forever()

    _loop_thread = threading.Thread(target=_run_loop, daemon=True)
    _loop_thread.start()
    return _loop


def run_async(func: Callable[..., Any], *args, **kwargs):
    return asyncio.to_thread(func, *args, **kwargs)


def run_sync(afunc: Callable[..., Any], *args, **kwargs):
    loop = _ensure_loop()

    coro = afunc(*args, **kwargs)
    future = asyncio.run_coroutine_threadsafe(coro, loop)
    return future.result()



from ._arg_parser import ArgParser
from ._component import Component
from ._context import Context, RunContext
from ._data_accessor import DataAccessor
from ._decorators import component, operation
from ._dot_dict import DotDict
from ._loader import Loader
from ._log_helper import warn
from ._ncall import NCall
from ._operation import Operation
from ._operation_parser import OperationParser
from ._provider import Provider
from ._response import Response
from ._type_converter import TypeConverter
from .data_model import DataModel

__all__ = [
    "ArgParser",
    "Component",
    "ComponentContext",
    "Context",
    "DataModel",
    "DataAccessor",
    "DotDict",
    "Loader",
    "NCall",
    "Operation",
    "OperationParser",
    "Provider",
    "ProviderContext",
    "RunContext",
    "Response",
    "TypeConverter",
    "component",
    "operation",
    "warn",
]



import inspect
import json
from typing import Any, get_args, get_origin, get_type_hints


class TypeConverter:
    @staticmethod
    def convert_value(value, expected_type):
        origin = get_origin(expected_type)

        # Handle Optional[T] (e.g., DataModel | None)
        if origin is not None and type(None) in get_args(expected_type):
            expected_type = next(
                t for t in get_args(expected_type) if t is not type(None)
            )
            origin = get_origin(expected_type)

        # Convert lists recursively
        if isinstance(value, list) and origin in (list, tuple):
            elem_type = (
                get_args(expected_type)[0] if get_args(expected_type) else Any
            )
            return [TypeConverter.convert_value(v, elem_type) for v in value]

        # Convert dictionaries recursively
        if isinstance(value, dict) and origin is dict:
            key_type, val_type = (
                get_args(expected_type)
                if get_args(expected_type)
                else (Any, Any)
            )
            return {
                TypeConverter.convert_value(
                    k, key_type
                ): TypeConverter.convert_value(v, val_type)
                for k, v in value.items()
            }

        # Convert dictionary to DataModel
        if (
            isinstance(value, dict)
            and hasattr(expected_type, "from_dict")
            and callable(getattr(expected_type, "from_dict"))
        ):
            return expected_type.from_dict(value)

        if (
            isinstance(value, str)
            and hasattr(expected_type, "from_dict")
            and callable(getattr(expected_type, "from_dict"))
        ):
            return expected_type.from_dict(json.loads(value))

        # Primitive type conversions
        try:
            if expected_type is int and isinstance(value, (str, float)):
                return int(value)
            if expected_type is float and isinstance(value, (str, int)):
                return float(value)
            if expected_type is str and isinstance(value, (int, float, bytes)):
                return str(value)
            if expected_type is bytes and isinstance(value, str):
                return value.encode()
            if expected_type is bool and isinstance(value, (str, int)):
                return (
                    bool(int(value))
                    if isinstance(value, str) and value.isdigit()
                    else bool(value)
                )
            if expected_type is list and not isinstance(value, list):
                return list(value)
            if expected_type is dict and isinstance(value, str):
                return json.loads(value)
            if expected_type is dict and isinstance(value, bytes):
                return json.loads(value.decode())
        except (ValueError, TypeError):
            pass  # If conversion fails, fallback to returning value as is.

        return value

    @staticmethod
    def convert_args(method, args: dict) -> dict:
        sig = inspect.signature(method)
        converted_args: dict = {}

        hints = get_type_hints(method)
        for param_name, _ in sig.parameters.items():
            param_type = hints.get(param_name, None)
            if param_name in args:
                converted_args[param_name] = TypeConverter.convert_value(
                    args[param_name], param_type
                )

        return args | converted_args



from typing import Any, Optional


class DotDict(dict):
    def __setattr__(self, name: str, value: Any) -> None:
        return super().__setitem__(name, value)

    def __getattr__(self, name: str) -> Any:
        try:
            return super().__getitem__(name)
        except KeyError:
            raise AttributeError(name)

    def __delattr__(self, name: str) -> None:
        return super().__delitem__(name)

    def __init__(self, dict: Optional[dict] = None):
        if dict is not None:
            for key, value in dict.items():
                if hasattr(value, "keys"):
                    value = DotDict(value)
                self[key] = value



from typing import Literal

from ._models import Field, Function, GeoPoint, Value

TextSearchQueryType = Literal[
    "simple",
    "full",
    "phrase",
    "phrase_prefix",
    "prefix",
    "best_fields",
    "most_fields",
    "cross_fields",
]

TextSearchMatchMode = Literal[
    "and",
    "or",
]


class QueryFunctionName:
    EXISTS = "exists"
    NOT_EXISTS = "not_exists"

    IS_DEFINED = "is_defined"
    IS_NOT_DEFINED = "is_not_defined"
    IS_TYPE = "is_type"

    LENGTH = "length"
    CONTAINS = "contains"
    STARTS_WITH = "starts_with"
    ENDS_WITH = "ends_with"

    ARRAY_LENGTH = "array_length"
    ARRAY_CONTAINS = "array_contains"
    ARRAY_CONTAINS_ANY = "array_contains_any"

    STARTS_WITH_DELIMITED = "starts_with_delimited"

    RANDOM = "random"
    NOW = "now"

    VECTOR_SEARCH = "vector_search"
    SPARSE_VECTOR_SEARCH = "sparse_vector_search"
    HYBRID_VECTOR_SEARCH = "hybrid_vector_search"
    TEXT_SEARCH = "text_search"
    HYBRID_TEXT_SEARCH = "hybrid_text_search"

    GEO_SEARCH_DISTANCE = "geo_search_distance"
    GEO_SEARCH_POLYGON = "geo_search_polygon"
    GEO_SEARCH_BBOX = "geo_search_bbox"


class QueryFunction:
    @staticmethod
    def exists() -> Function:
        return Function(name=QueryFunctionName.EXISTS)

    @staticmethod
    def not_exists() -> Function:
        return Function(name=QueryFunctionName.NOT_EXISTS)

    @staticmethod
    def is_defined(field: str) -> Function:
        return Function(
            name=QueryFunctionName.IS_DEFINED, args=[Field(path=field)]
        )

    @staticmethod
    def is_not_defined(field: str) -> Function:
        return Function(
            name=QueryFunctionName.IS_NOT_DEFINED, args=[Field(path=field)]
        )

    @staticmethod
    def is_type(field: str, type: str) -> Function:
        return Function(
            name=QueryFunctionName.IS_TYPE,
            args=[
                Field(path=field),
                type,
            ],
        )

    @staticmethod
    def starts_with(field: str, value: str) -> Function:
        return Function(
            name=QueryFunctionName.STARTS_WITH,
            args=[
                Field(path=field),
                value,
            ],
        )

    @staticmethod
    def ends_with(field: str, value: str) -> Function:
        return Function(
            name=QueryFunctionName.ENDS_WITH,
            args=[
                Field(path=field),
                value,
            ],
        )

    @staticmethod
    def contains(field: str, value: str) -> Function:
        return Function(
            name=QueryFunctionName.CONTAINS,
            args=[
                Field(path=field),
                value,
            ],
        )

    @staticmethod
    def length(field: str) -> Function:
        return Function(
            name=QueryFunctionName.LENGTH, args=[Field(path=field)]
        )

    @staticmethod
    def array_contains(field: str, value: Value) -> Function:
        return Function(
            name=QueryFunctionName.ARRAY_CONTAINS,
            args=[
                Field(path=field),
                value,
            ],
        )

    @staticmethod
    def array_contains_any(field: str, value: list[Value]) -> Function:
        return Function(
            name=QueryFunctionName.ARRAY_CONTAINS_ANY,
            args=[
                Field(path=field),
                value,
            ],
        )

    @staticmethod
    def array_length(field: str) -> Function:
        return Function(
            name=QueryFunctionName.ARRAY_LENGTH, args=[Field(path=field)]
        )

    @staticmethod
    def starts_with_delimited(
        field: str,
        value: str,
        delimiter: str,
    ) -> Function:
        return Function(
            name=QueryFunctionName.STARTS_WITH_DELIMITED,
            args=[Field(path=field), value, delimiter],
        )

    @staticmethod
    def random() -> Function:
        return Function(name=QueryFunctionName.RANDOM)

    @staticmethod
    def now() -> Function:
        return Function(name=QueryFunctionName.NOW)

    @staticmethod
    def vector_search_old(
        vector: list[float] | None = None,
        sparse_vector: dict[int, float] | None = None,
        field: str | None = None,
    ):
        return Function(
            name=QueryFunctionName.VECTOR_SEARCH,
            named_args=dict(
                vector=vector, sparse_vector=sparse_vector, field=field
            ),
        )

    @staticmethod
    def vector_search(
        vector: list[float] | None = None,
        field: str | None = None,
        k: int | None = None,
        num_candidates: int | None = None,
    ) -> Function:
        return Function(
            name=QueryFunctionName.VECTOR_SEARCH,
            named_args=dict(
                vector=vector,
                field=field,
                k=k,
                num_candidates=num_candidates,
            ),
        )

    @staticmethod
    def sparse_vector_search(
        sparse_vector: dict[int, float] | None = None, field: str | None = None
    ) -> Function:
        return Function(
            name=QueryFunctionName.SPARSE_VECTOR_SEARCH,
            named_args=dict(
                sparse_vector=sparse_vector,
                field=field,
            ),
        )

    @staticmethod
    def hybrid_vector_search(
        # dense vector
        vector: list[float] | None = None,
        vector_field: str | None = None,
        k: int | None = None,
        num_candidates: int | None = None,
        # sparse vector
        sparse_vector: dict[int, float] | None = None,
        sparse_vector_field: str | None = None,
        # fusion
        hybrid_mode: (
            Literal["add", "multiply", "max", "min", "rrf"] | None
        ) = None,
        vector_weight: float | None = None,
        sparse_vector_weight: float | None = None,
    ) -> Function:
        return Function(
            name=QueryFunctionName.HYBRID_VECTOR_SEARCH,
            named_args=dict(
                # dense
                vector=vector,
                vector_field=vector_field,
                k=k,
                num_candidates=num_candidates,
                # sparse
                sparse_vector=sparse_vector,
                sparse_vector_field=sparse_vector_field,
                # fusion
                hybrid_mode=hybrid_mode,
                vector_weight=vector_weight,
                sparse_vector_weight=sparse_vector_weight,
            ),
        )

    @staticmethod
    def text_search(
        query: str,
        fields: list[str] | None = None,
        match_mode: TextSearchMatchMode | None = None,
        query_type: TextSearchQueryType | None = None,
        fuzziness: int | str | None = None,
        minimum_should_match: int | str | None = None,
        analyzer: str | None = None,
        boost: dict[str, float] | None = None,
    ) -> Function:
        return Function(
            name=QueryFunctionName.TEXT_SEARCH,
            named_args=dict(
                query=query,
                fields=fields,
                match_mode=match_mode,
                query_type=query_type,
                fuzziness=fuzziness,
                minimum_should_match=minimum_should_match,
                analyzer=analyzer,
                boost=boost,
            ),
        )

    @staticmethod
    def hybrid_text_search(
        # text
        query: str | None = None,
        fields: list[str] | None = None,
        match_mode: TextSearchMatchMode | None = None,
        query_type: TextSearchQueryType | None = None,
        fuzziness: int | str | None = None,
        minimum_should_match: int | str | None = None,
        analyzer: str | None = None,
        boost: dict[str, float] | None = None,
        # vector
        vector: list[float] | None = None,
        vector_field: str | None = None,
        k: int | None = None,
        num_candidates: int | None = None,
        # sparse vector
        sparse_vector: dict[int, float] | None = None,
        sparse_vector_field: str | None = None,
        # fusion
        hybrid_mode: (
            Literal["add", "multiply", "max", "min", "rrf"] | None
        ) = None,
        text_weight: float | None = None,
        vector_weight: float | None = None,
    ) -> Function:
        return Function(
            name=QueryFunctionName.HYBRID_TEXT_SEARCH,
            named_args=dict(
                query=query,
                fields=fields,
                term_match_mode=match_mode,
                query_type=query_type,
                fuzziness=fuzziness,
                minimum_should_match=minimum_should_match,
                analyzer=analyzer,
                boost=boost,
                vector=vector,
                vector_field=vector_field,
                k=k,
                num_candidates=num_candidates,
                sparse_vector=sparse_vector,
                sparse_vector_field=sparse_vector_field,
                hybrid_mode=hybrid_mode,
                text_weight=text_weight,
                vector_weight=vector_weight,
            ),
        )

    @staticmethod
    def geo_search_distance(
        field: str,
        center: dict | GeoPoint,
        distance: float,
        unit: Literal["m", "km", "mi"] = "m",
    ) -> Function:
        return Function(
            name=QueryFunctionName.GEO_SEARCH_DISTANCE,
            named_args=dict(
                field=Field(path=field),
                center=(
                    center
                    if isinstance(center, dict)
                    else {"lat": center.lat, "lon": center.lon}
                ),
                distance=distance,
                unit=unit,
            ),
        )

    @staticmethod
    def geo_search_polygon(
        field: str,
        points: list[dict] | list[GeoPoint],
    ) -> Function:
        return Function(
            name=QueryFunctionName.GEO_SEARCH_POLYGON,
            named_args=dict(
                field=Field(path=field),
                points=[
                    p if isinstance(p, dict) else {"lat": p.lat, "lon": p.lon}
                    for p in points
                ],
            ),
        )

    @staticmethod
    def geo_search_bbox(
        field: str,
        top_left: dict | GeoPoint,
        bottom_right: dict | GeoPoint,
    ) -> Function:
        return Function(
            name=QueryFunctionName.GEO_SEARCH_BBOX,
            named_args=dict(
                field=Field(path=field),
                top_left=(
                    top_left
                    if isinstance(top_left, dict)
                    else {"lat": top_left.lat, "lon": top_left.lon}
                ),
                bottom_right=(
                    bottom_right
                    if isinstance(bottom_right, dict)
                    else {
                        "lat": bottom_right.lat,
                        "lon": bottom_right.lon,
                    }
                ),
            ),
        )



__all__ = ["DataModel", "DataModelField", "Empty", "get_origin", "get_args"]

from typing import Any, Self

from pydantic import BaseModel, ConfigDict, Field
from pydantic._internal._generics import get_args as get_pydantic_get_args
from pydantic._internal._generics import get_origin as get_pydantic_origin
from pydantic_core import PydanticUndefined


def get_origin(v: Any):
    return get_pydantic_origin(v)


def get_args(v: Any):
    return get_pydantic_get_args(v)


class DataModel(BaseModel):
    """Data model."""

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="ignore")

    def to_dict(self):
        return self.model_dump()

    def to_json(self, indent: int | None = None):
        return self.model_dump_json(indent=indent)

    def copy(self, deep: bool = False, **kwargs):
        return self.model_copy(deep=deep, **kwargs)

    @classmethod
    def from_dict(cls, obj: dict | None) -> Self:
        return cls.model_validate(obj)

    @classmethod
    def from_json(cls, json: str) -> Self:
        return cls.model_validate_json(json)

    @classmethod
    def get_default_value(cls, name: str) -> Any:
        field = cls.model_fields.get(name)
        if field:
            if field.default is PydanticUndefined:
                return Empty
            return field.default
        raise ValueError(f"Attribute {name} not found in model")


class Empty(DataModel):
    """Empty value."""

    pass


def DataModelField(
    alias: str | None = None,
    exclude: bool | None = None,
    min_length: int | None = None,
    max_length: int | None = None,
    **kwargs,
) -> Any:
    return Field(
        alias=alias,
        exclude=exclude,
        min_length=min_length,
        max_length=max_length,
        **kwargs,
    )



from __future__ import annotations

from typing import Any

from ..ql._functions import QueryFunctionName
from ..ql._models import (
    Collection,
    Expression,
    Function,
    FunctionNamespace,
    Not,
    OrderBy,
    Select,
    Update,
)
from ._arg_parser import ArgParser
from ._operation import Operation
from .data_model import DataModel


class OperationParser:
    operation: Operation | None

    def __init__(self, operation: Operation | None):
        self.operation = operation

    def op_equals(self, name: str | None) -> bool:
        operation_name = self.get_op_name()
        if operation_name is not None:
            operation_name = operation_name.lower()
        arg_name = name
        if arg_name is not None:
            arg_name = arg_name.lower()
        return operation_name == arg_name

    def get_args(self) -> dict:
        if self.operation is not None and self.operation.args is not None:
            return self.operation.args
        return {}

    def get_nargs(self) -> dict | None:
        return self.get_arg("nargs")

    def get_arg(self, arg: str, convert_dict: bool = False) -> Any:
        if (
            self.operation is not None
            and self.operation.args is not None
            and arg in self.operation.args
        ):
            arg_val = self.operation.args[arg]
            if convert_dict and isinstance(arg_val, DataModel):
                return arg_val.to_dict()
            return arg_val
        return None

    def set_arg(self, arg: str, value: Any) -> None:
        if self.operation is not None:
            if self.operation.args is None:
                self.operation.args = dict()
            self.operation.args[arg] = value

    def has_arg(self, arg: str) -> bool:
        return (
            self.operation is not None
            and self.operation.args is not None
            and arg in self.operation.args
        )

    def get_op_name(self) -> str | None:
        if self.operation is not None:
            return self.operation.name
        return None

    def get_statement(self) -> str | None:
        return self.get_arg("statement")

    @staticmethod
    def parse_select(
        select: str | Select | None,
        params: dict[str, Any] | None = None,
    ) -> Select | None:
        return ArgParser.get_parsed_arg(
            "select",
            select,
            params,
        )

    def get_select(self) -> Select | None:
        return OperationParser.parse_select(
            self.get_arg("select"),
            self.get_params(),
        )

    @staticmethod
    def parse_collection_name(
        collection: str | Collection | None = None,
        params: dict[str, Any] | None = None,
    ) -> str | None:
        col = ArgParser.get_parsed_arg(
            "collection",
            collection,
            params,
        )
        if col is None:
            return None
        return col.name

    def get_collection_name(self) -> str | None:
        return OperationParser.parse_collection_name(
            self.get_arg("collection"),
            self.get_params(),
        )

    @staticmethod
    def parse_set(
        set: str | Update | None,
        params: dict[str, Any] | None = None,
    ) -> Update:
        return ArgParser.get_parsed_arg(
            "set",
            set,
            params,
        )

    def get_set(self) -> Update:
        return OperationParser.parse_set(
            self.get_arg("set"),
            self.get_params(),
        )

    @staticmethod
    def parse_search(
        search: str | Expression | None,
        params: dict[str, Any] | None = None,
    ) -> Select | None:
        return ArgParser.get_parsed_arg(
            "search",
            search,
            params,
        )

    def get_search(self) -> Expression | None:
        return OperationParser.parse_search(
            self.get_arg("search"),
            self.get_params(),
        )

    @staticmethod
    def parse_where(
        where: str | Expression | None,
        params: dict[str, Any] | None = None,
    ) -> Expression | None:
        return ArgParser.get_parsed_arg(
            "where",
            where,
            params,
        )

    def get_where(self) -> Expression | None:
        return OperationParser.parse_where(
            self.get_arg("where"),
            self.get_params(),
        )

    @staticmethod
    def parse_order_by(
        order_by: str | OrderBy | None,
        params: dict[str, Any] | None = None,
    ) -> OrderBy | None:
        return ArgParser.get_parsed_arg(
            "order_by",
            order_by,
            params,
        )

    def get_order_by(self) -> OrderBy | None:
        return OperationParser.parse_order_by(
            self.get_arg("order_by"),
            self.get_params(),
        )

    @staticmethod
    def parse_rank_by(
        rank_by: str | Expression | None,
        params: dict[str, Any] | None = None,
    ) -> Expression | None:
        return ArgParser.get_parsed_arg(
            "rank_by",
            rank_by,
            params,
        )

    def get_rank_by(self) -> Expression | None:
        return OperationParser.parse_rank_by(
            self.get_arg("rank_by"),
            self.get_params(),
        )

    def get_limit(self) -> int | None:
        return self.get_arg("limit")

    def get_offset(self) -> int | None:
        return self.get_arg("offset")

    def get_returning(self) -> str | None:
        return self.get_arg("returning")

    def get_returning_as_bool(self) -> bool | None:
        returning = self.get_arg("returning")
        if returning:
            return returning == "new"
        return None

    def get_continuation(self) -> str | None:
        return self.get_arg("continuation")

    def get_config(self) -> Any:
        return self.get_arg("config")

    def get_params(self) -> dict[str, Any] | None:
        return self.get_arg("params")

    def get_nflags(self) -> dict | None:
        return self.get_arg("nflags")

    @staticmethod
    def parse_where_exists(
        where: str | Expression | None,
        params: dict[str, Any] | None = None,
    ) -> bool | None:
        expr = OperationParser.parse_where(where, params)
        if expr is not None:
            exists = True
            if isinstance(expr, Not):
                expr = expr.expr
                exists = False
            if isinstance(expr, Function):
                if (
                    expr.namespace == FunctionNamespace.BUILTIN
                    and expr.name == QueryFunctionName.EXISTS
                ):
                    return exists
                if (
                    expr.namespace == FunctionNamespace.BUILTIN
                    and expr.name == QueryFunctionName.NOT_EXISTS
                ):
                    return not exists
        return None

    def get_where_exists(self) -> bool | None:
        return OperationParser.parse_where_exists(
            self.get_arg("where"),
            self.get_params(),
        )



from __future__ import annotations

import uuid
from typing import Any

from ._async_helper import run_async, run_sync
from ._context import Context
from ._operation import Operation
from ._provider import Provider
from ._response import Response
from ._type_converter import TypeConverter
from .exceptions import NotSupportedError


class Component:
    __provider__: Provider
    __handle__: str | None
    __type__: str
    __unpack__: bool
    __native__: bool

    def __init__(
        self,
        **kwargs,
    ):
        self.__native__ = kwargs.pop("__native__", False)
        self.__unpack__ = kwargs.pop("__unpack__", False)
        self.__handle__ = kwargs.pop("__handle__", None)
        self.__type__ = kwargs.pop("__type__", self.__class__.__module__)
        if "__provider__" in kwargs:
            self.__bind__(kwargs.pop("__provider__"))

    def __bind__(
        self,
        provider: Provider | dict | str | None,
    ) -> None:
        if provider is None:
            return
        if isinstance(provider, Provider):
            provider.__component__ = self
            self.__provider__ = provider
        else:
            if isinstance(provider, dict):
                type = provider.pop("type")
                parameters = provider.pop("parameters", dict())
            elif isinstance(provider, str):
                type = provider
                parameters = dict()
            from ._loader import Loader

            module_name = self.__class__.__module__.rsplit(".", 1)[0]
            provider_path = f"{module_name}.providers.{type}"
            provider_instance = Loader.load_provider_instance(
                path=provider_path,
                parameters=parameters,
            )
            self.__bind__(provider=provider_instance)

    def __setup__(self, context: Context | None = None) -> None:
        self.__provider__.__setup__(context=context)

    async def __asetup__(self, context: Context | None = None) -> None:
        await self.__provider__.__asetup__(context=context)

    def __run__(
        self,
        operation: dict | str | Operation | None = None,
        context: dict | Context | None = None,
        **kwargs,
    ) -> Any:
        current_context = self._init_context(context)
        if hasattr(self, "__provider__"):
            response = self.__provider__.__run__(
                operation=self._convert_operation(operation),
                context=current_context,
                **kwargs,
            )
        elif operation and operation.name:
            func_name = operation.name
            func = getattr(self, func_name, None)
            if func and callable(func):
                args = TypeConverter.convert_args(func, operation.args or {})
                response = func(**args)
            else:
                afunc_name = f"a{operation.name}"
                afunc = getattr(self, afunc_name, None)
                if callable(afunc):
                    args = TypeConverter.convert_args(
                        afunc, operation.args or {}
                    )
                    response = run_sync(afunc, **args)
                else:
                    raise NotSupportedError(
                        operation.to_json() if operation is not None else None
                    )
        else:
            raise NotSupportedError()

        if (
            hasattr(self, "__native__")
            and not self.__native__
            and isinstance(response, Response)
        ):
            response.native = None
        if (
            hasattr(self, "__unpack__")
            and self.__unpack__
            and isinstance(response, Response)
        ):
            return response.result
        return response

    async def __arun__(
        self,
        operation: dict | str | Operation | None = None,
        context: dict | Context | None = None,
        **kwargs,
    ) -> Any:
        current_context = self._init_context(context)
        if hasattr(self, "__provider__"):
            response = await self.__provider__.__arun__(
                operation=self._convert_operation(operation),
                context=current_context,
                **kwargs,
            )
        else:
            afunc_name = f"a{operation.name}"
            afunc = getattr(self, afunc_name, None)
            if afunc and callable(afunc):
                args = TypeConverter.convert_args(afunc, operation.args or {})
                response = await afunc(**args)
            else:
                response = await run_async(
                    func=self.__run__,
                    operation=operation,
                    context=context,
                    **kwargs,
                )

        if (
            hasattr(self, "__native__")
            and not self.__native__
            and isinstance(response, Response)
        ):
            response.native = None
        if (
            hasattr(self, "__unpack__")
            and self.__unpack__
            and isinstance(response, Response)
        ):
            return response.result
        return response

    def __supports__(self, feature: str) -> bool:
        return self.__provider__.__supports__(feature)

    def __execute__(
        self,
        statement: str,
        params: dict[str, Any] | None = None,
        context: Context | None = None,
        **kwargs: Any,
    ) -> Any:
        return self.__provider__.__execute__(
            statement, params, context, **kwargs
        )

    async def __aexecute__(
        self,
        statement: str,
        params: dict[str, Any] | None = None,
        context: Context | None = None,
        **kwargs: Any,
    ) -> Any:
        return await self.__provider__.__aexecute__(
            statement, params, context, **kwargs
        )

    def __serialize__(self) -> Any:
        pass

    def _convert_operation(
        self,
        operation: dict | str | Operation | None,
    ) -> Operation | None:
        if isinstance(operation, dict):
            return Operation.from_dict(operation)
        elif isinstance(operation, str):
            return Operation(name=operation)
        return operation

    def _init_context(
        self,
        context: dict | Context | None,
    ) -> Context:
        if isinstance(context, dict):
            context = Context.from_dict(context)
        if context is not None:
            id = context.id
        else:
            id = str(uuid.uuid4())
        ctx = Context(
            id=id,
            data=context.data if context else None,
        )
        return ctx



import copy
from functools import lru_cache
from typing import Any

from antlr4 import CommonTokenStream  # type: ignore
from antlr4 import InputStream  # type: ignore
from antlr4 import ParseTreeWalker  # type: ignore
from antlr4.error.ErrorListener import ConsoleErrorListener, ErrorListener

from ..core._operation import Operation
from ._models import Collection, Expression, OrderBy, Select, Update
from ._parser_listener import VerseQLParserListener  # type: ignore
from .exceptions import ParserError
from .generated.VerseQLLexer import VerseQLLexer  # type: ignore
from .generated.VerseQLParser import VerseQLParser  # type: ignore


class QLParser:
    @lru_cache
    @staticmethod
    def parse(str: str, type: str) -> Any:
        obj = None
        if str is None:
            return None
        parser = QLParser._get_parser(str=str)
        if type == "statement":
            tree = parser.parse_statement()
            listener = QLParser._get_listener(tree)
            obj = listener.operation
        elif type == "where":
            tree = parser.parse_where()
            listener = QLParser._get_listener(tree)
            obj = listener.where
        elif type == "select":
            tree = parser.parse_select()
            listener = QLParser._get_listener(tree)
            obj = listener.select
        elif type == "collection":
            tree = parser.parse_collection()
            listener = QLParser._get_listener(tree)
            obj = listener.collection
        elif type == "order_by":
            tree = parser.parse_order_by()
            listener = QLParser._get_listener(tree)
            obj = listener.order_by
        elif type == "rank_by":
            tree = parser.parse_rank_by()
            listener = QLParser._get_listener(tree)
            obj = listener.rank_by
        elif type == "search":
            tree = parser.parse_search()
            listener = QLParser._get_listener(tree)
            obj = listener.search
        elif type == "update":
            tree = parser.parse_update()
            listener = QLParser._get_listener(tree)
            obj = listener.update
        return obj

    @staticmethod
    def parse_statement(str: str) -> Operation | None:
        return copy.deepcopy(QLParser.parse(str, "statement"))

    @staticmethod
    def parse_where(str: str) -> Expression:
        return copy.deepcopy(QLParser.parse(str, "where"))

    @staticmethod
    def parse_select(str: str) -> Select | None:
        return copy.deepcopy(QLParser.parse(str, "select"))

    @staticmethod
    def parse_collection(str: str) -> Collection | None:
        return copy.deepcopy(QLParser.parse(str, "collection"))

    @staticmethod
    def parse_order_by(str: str) -> OrderBy | None:
        return copy.deepcopy(QLParser.parse(str, "order_by"))

    @staticmethod
    def parse_rank_by(str: str) -> Expression | None:
        return copy.deepcopy(QLParser.parse(str, "rank_by"))

    @staticmethod
    def parse_search(str: str) -> Expression:
        return copy.deepcopy(QLParser.parse(str, "search"))

    @staticmethod
    def parse_update(str: str) -> Update | None:
        return copy.deepcopy(QLParser.parse(str, "update"))

    @staticmethod
    def _get_listener(tree) -> VerseQLParserListener:
        listener = VerseQLParserListener()
        walker = ParseTreeWalker()
        walker.walk(listener, tree)
        return listener

    @staticmethod
    def _get_parser(str: str) -> VerseQLParser:
        input_stream = InputStream(str)
        lexer = VerseQLLexer(input_stream)
        stream = CommonTokenStream(lexer)
        parser = VerseQLParser(stream)
        parser.removeErrorListener(ConsoleErrorListener.INSTANCE)
        parser.addErrorListener(ParserErrorListener())
        return parser


class ParserErrorListener(ErrorListener):
    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        raise ParserError("line " + str(line) + ":" + str(column) + " " + msg)

    def reportAmbiguity(
        self, recognizer, dfa, startIndex, stopIndex, exact, ambigAlts, configs
    ):
        pass

    def reportAttemptingFullContext(
        self, recognizer, dfa, startIndex, stopIndex, conflictingAlts, configs
    ):
        pass

    def reportContextSensitivity(
        self, recognizer, dfa, startIndex, stopIndex, prediction, configs
    ):
        pass



from ._functions import (
    QueryFunction,
    QueryFunctionName,
    TextSearchMatchMode,
    TextSearchQueryType,
)
from ._models import (
    And,
    Collection,
    Comparison,
    ComparisonOp,
    Expression,
    Field,
    Function,
    FunctionNamespace,
    GeoPoint,
    Not,
    Or,
    OrderBy,
    OrderByDirection,
    OrderByTerm,
    Parameter,
    Ref,
    Select,
    SelectTerm,
    Undefined,
    Update,
    UpdateOp,
    UpdateOperation,
    Value,
)
from ._ql_parser import QLParser
from ._query_processor import QueryProcessor

__all__ = [
    "And",
    "Collection",
    "Comparison",
    "ComparisonOp",
    "Expression",
    "Expression",
    "Field",
    "Function",
    "FunctionNamespace",
    "Not",
    "Or",
    "OrderBy",
    "OrderByDirection",
    "OrderByTerm",
    "Parameter",
    "Ref",
    "Select",
    "SelectTerm",
    "Undefined",
    "Update",
    "UpdateOp",
    "UpdateOperation",
    "Value",
    "QLParser",
    "QueryFunction",
    "QueryFunctionName",
    "QueryProcessor",
    "TextSearchMatchMode",
    "TextSearchQueryType",
    "GeoPoint",
]



from __future__ import annotations

from typing import Any

from ._yaml_loader import YamlLoader
from .data_model import DataModel

__all__ = [
    "ComponentConfig",
    "Manifest",
    "ProviderConfig",
    "MANIFEST_FILE",
    "ManifestMetdata",
]


MANIFEST_FILE = "verse.yaml"


class ManifestMetdata:
    name: str | None = None
    description: str | None = None
    version: str | None = None


class ProviderConfig(DataModel):
    type: str
    parameters: dict[str, Any] = dict()
    requirements: list[str] = list()


class ComponentConfig(DataModel):
    type: str
    parameters: dict[str, Any] = dict()
    providers: dict[str, ProviderConfig] = dict()
    requirements: list[str] = list()


class Manifest(DataModel):
    metadata: ManifestMetdata = ManifestMetdata()
    variables: dict[str, Any] = dict()
    components: dict[str, ComponentConfig] = dict()
    bindings: dict[str, dict[str, str]] = dict()
    requirements: list[str] = list()

    @staticmethod
    def parse(path: str) -> Manifest:
        obj = YamlLoader.load(path=path)
        manifest = Manifest.from_dict(obj)
        return manifest



# flake8: noqa
# type: ignore
import decimal
import json

from antlr4 import *

from ..core._operation import Operation
from ._models import (
    And,
    Collection,
    Comparison,
    ComparisonOp,
    Field,
    Function,
    Not,
    Or,
    OrderBy,
    OrderByTerm,
    Parameter,
    Ref,
    Select,
    SelectTerm,
    Update,
    UpdateOperation,
)
from .generated.VerseQLParser import VerseQLParser


# This class defines a complete listener for a parse tree produced by VerseQLParser.
class VerseQLParserListener(ParseTreeListener):
    operation = None
    select = None
    collection = None
    where = None
    order_by = None
    update = None
    search = None
    rank_by = None

    # Enter a parse tree produced by VerseQLParser#identifier.
    def enterIdentifier(self, ctx: VerseQLParser.IdentifierContext):
        pass

    # Exit a parse tree produced by VerseQLParser#identifier.
    def exitIdentifier(self, ctx: VerseQLParser.IdentifierContext):
        ctx.text = ctx.getText()

    # Enter a parse tree produced by VerseQLParser#parse_statement.
    def enterParse_statement(self, ctx: VerseQLParser.Parse_statementContext):
        pass

    # Exit a parse tree produced by VerseQLParser#parse_statement.
    def exitParse_statement(self, ctx: VerseQLParser.Parse_statementContext):
        self.operation = ctx.statement().val

    # Enter a parse tree produced by VerseQLParser#parse_search.
    def enterParse_search(self, ctx: VerseQLParser.Parse_searchContext):
        pass

    # Exit a parse tree produced by VerseQLParser#parse_search.
    def exitParse_search(self, ctx: VerseQLParser.Parse_searchContext):
        self.search = ctx.expression().val

    # Enter a parse tree produced by VerseQLParser#parse_where.
    def enterParse_where(self, ctx: VerseQLParser.Parse_whereContext):
        pass

    # Exit a parse tree produced by VerseQLParser#parse_where.
    def exitParse_where(self, ctx: VerseQLParser.Parse_whereContext):
        self.where = ctx.expression().val

    # Enter a parse tree produced by VerseQLParser#parse_select.
    def enterParse_select(self, ctx: VerseQLParser.Parse_selectContext):
        pass

    # Exit a parse tree produced by VerseQLParser#parse_select.
    def exitParse_select(self, ctx: VerseQLParser.Parse_selectContext):
        self.select = ctx.select().val

    # Enter a parse tree produced by VerseQLParser#parse_collection.
    def enterParse_collection(
        self, ctx: VerseQLParser.Parse_collectionContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#parse_collection.
    def exitParse_collection(self, ctx: VerseQLParser.Parse_collectionContext):
        self.collection = ctx.collection().val

    # Enter a parse tree produced by VerseQLParser#parse_order_by.
    def enterParse_order_by(self, ctx: VerseQLParser.Parse_order_byContext):
        pass

    # Exit a parse tree produced by VerseQLParser#parse_order_by.
    def exitParse_order_by(self, ctx: VerseQLParser.Parse_order_byContext):
        self.order_by = ctx.order_by().val

    # Enter a parse tree produced by VerseQLParser#parse_rank_by.
    def enterParse_rank_by(self, ctx: VerseQLParser.Parse_rank_byContext):
        pass

    # Exit a parse tree produced by VerseQLParser#parse_rank_by.
    def exitParse_rank_by(self, ctx: VerseQLParser.Parse_rank_byContext):
        self.rank_by = ctx.expression().val

    # Enter a parse tree produced by VerseQLParser#parse_update.
    def enterParse_update(self, ctx: VerseQLParser.Parse_updateContext):
        pass

    # Exit a parse tree produced by VerseQLParser#parse_update.
    def exitParse_update(self, ctx: VerseQLParser.Parse_updateContext):
        self.update = ctx.update().val

    # Enter a parse tree produced by VerseQLParser#statement_single.
    def enterStatement_single(
        self, ctx: VerseQLParser.Statement_singleContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#statement_single.
    def exitStatement_single(self, ctx: VerseQLParser.Statement_singleContext):
        op = ctx.op.text.lower()
        args = dict()
        for clause in ctx.clause():
            args[clause.key] = clause.val
        ctx.val = Operation(name=op, args=args)

    # Enter a parse tree produced by VerseQLParser#statement_multi.
    def enterStatement_multi(self, ctx: VerseQLParser.Statement_multiContext):
        pass

    # Exit a parse tree produced by VerseQLParser#statement_multi.
    def exitStatement_multi(self, ctx: VerseQLParser.Statement_multiContext):
        op = ctx.op.text.lower()
        operations = []
        for statement in ctx.statement():
            operations.append(statement.val)
        if op == "batch":
            arg = "batch"
        elif op == "transact":
            arg = "transaction"
        else:
            arg = op
        ctx.val = Operation(name=op, args={arg: {"operations": operations}})

    # Enter a parse tree produced by VerseQLParser#clause.
    def enterClause(self, ctx: VerseQLParser.ClauseContext):
        pass

    # Exit a parse tree produced by VerseQLParser#clause.
    def exitClause(self, ctx: VerseQLParser.ClauseContext):
        clause = None
        if ctx.select_clause() is not None:
            clause = ctx.select_clause()
        elif ctx.collection_clause() is not None:
            clause = ctx.collection_clause()
        elif ctx.set_clause() is not None:
            clause = ctx.set_clause()
        elif ctx.where_clause() is not None:
            clause = ctx.where_clause()
        elif ctx.order_by_clause() is not None:
            clause = ctx.order_by_clause()
        elif ctx.rank_by_clause() is not None:
            clause = ctx.rank_by_clause()
        elif ctx.search_clause() is not None:
            clause = ctx.search_clause()
        elif ctx.generic_clause() is not None:
            clause = ctx.generic_clause()
        ctx.key = clause.key
        ctx.val = clause.val

    # Enter a parse tree produced by VerseQLParser#generic_clause.
    def enterGeneric_clause(self, ctx: VerseQLParser.Generic_clauseContext):
        pass

    # Exit a parse tree produced by VerseQLParser#generic_clause.
    def exitGeneric_clause(self, ctx: VerseQLParser.Generic_clauseContext):
        ctx.key = ctx.name.text.lower()
        ctx.val = ctx.operand().val

    # Enter a parse tree produced by VerseQLParser#select_clause.
    def enterSelect_clause(self, ctx: VerseQLParser.Select_clauseContext):
        pass

    # Exit a parse tree produced by VerseQLParser#select_clause.
    def exitSelect_clause(self, ctx: VerseQLParser.Select_clauseContext):
        ctx.key = "select"
        ctx.val = ctx.select().val

    # Enter a parse tree produced by VerseQLParser#select_all.
    def enterSelect_all(self, ctx: VerseQLParser.Select_allContext):
        pass

    # Exit a parse tree produced by VerseQLParser#select_all.
    def exitSelect_all(self, ctx: VerseQLParser.Select_allContext):
        ctx.val = Select()

    # Enter a parse tree produced by VerseQLParser#select_terms.
    def enterSelect_terms(self, ctx: VerseQLParser.Select_termsContext):
        pass

    # Exit a parse tree produced by VerseQLParser#select_terms.
    def exitSelect_terms(self, ctx: VerseQLParser.Select_termsContext):
        ctx.val = Select(terms=[item.val for item in ctx.select_term()])

    # Enter a parse tree produced by VerseQLParser#select_parameter.
    def enterSelect_parameter(
        self, ctx: VerseQLParser.Select_parameterContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#select_parameter.
    def exitSelect_parameter(self, ctx: VerseQLParser.Select_parameterContext):
        ctx.val = ctx.parameter().val

    # Enter a parse tree produced by VerseQLParser#select_term.
    def enterSelect_term(self, ctx: VerseQLParser.Select_termContext):
        pass

    # Exit a parse tree produced by VerseQLParser#select_term.
    def exitSelect_term(self, ctx: VerseQLParser.Select_termContext):
        ctx.val = SelectTerm(
            field=ctx.field()[0].val.path,
            alias=None if len(ctx.field()) == 1 else ctx.field()[1].val.path,
        )

    # Enter a parse tree produced by VerseQLParser#collection_clause.
    def enterCollection_clause(
        self, ctx: VerseQLParser.Collection_clauseContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#collection_clause.
    def exitCollection_clause(
        self, ctx: VerseQLParser.Collection_clauseContext
    ):
        ctx.key = "collection"
        ctx.val = ctx.collection().val

    # Enter a parse tree produced by VerseQLParser#collection_identifier.
    def enterCollection_identifier(
        self, ctx: VerseQLParser.Collection_identifierContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#collection_identifier.
    def exitCollection_identifier(
        self, ctx: VerseQLParser.Collection_identifierContext
    ):
        ctx.val = Collection(name=ctx.getText())

    # Enter a parse tree produced by VerseQLParser#collection_parameter.
    def enterCollection_parameter(
        self, ctx: VerseQLParser.Collection_parameterContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#collection_parameter.
    def exitCollection_parameter(
        self, ctx: VerseQLParser.Collection_parameterContext
    ):
        ctx.val = ctx.parameter().val

    # Enter a parse tree produced by VerseQLParser#search_clause.
    def enterSearch_clause(self, ctx: VerseQLParser.Search_clauseContext):
        pass

    # Exit a parse tree produced by VerseQLParser#search_clause.
    def exitSearch_clause(self, ctx: VerseQLParser.Search_clauseContext):
        ctx.key = "search"
        ctx.val = ctx.expression().val

    # Enter a parse tree produced by VerseQLParser#where_clause.
    def enterWhere_clause(self, ctx: VerseQLParser.Where_clauseContext):
        pass

    # Exit a parse tree produced by VerseQLParser#where_clause.
    def exitWhere_clause(self, ctx: VerseQLParser.Where_clauseContext):
        ctx.key = "where"
        ctx.val = ctx.expression().val

    # Enter a parse tree produced by VerseQLParser#expression_operand.
    def enterExpression_operand(
        self, ctx: VerseQLParser.Expression_operandContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#expression_operand.
    def exitExpression_operand(
        self, ctx: VerseQLParser.Expression_operandContext
    ):
        ctx.val = ctx.operand().val

    # Enter a parse tree produced by VerseQLParser#expression_paranthesis.
    def enterExpression_paranthesis(
        self, ctx: VerseQLParser.Expression_paranthesisContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#expression_paranthesis.
    def exitExpression_paranthesis(
        self, ctx: VerseQLParser.Expression_paranthesisContext
    ):
        ctx.val = ctx.expression().val

    # Enter a parse tree produced by VerseQLParser#expression_comparison.
    def enterExpression_comparison(
        self, ctx: VerseQLParser.Expression_comparisonContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#expression_comparison.
    def exitExpression_comparison(
        self, ctx: VerseQLParser.Expression_comparisonContext
    ):
        ctx.val = Comparison(
            lexpr=ctx.lhs.val,
            op=ctx.op.text.lower(),
            rexpr=ctx.rhs.val,
        )

    # Enter a parse tree produced by VerseQLParser#expression_comparison_between.
    def enterExpression_comparison_between(
        self, ctx: VerseQLParser.Expression_comparison_betweenContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#expression_comparison_between.
    def exitExpression_comparison_between(
        self, ctx: VerseQLParser.Expression_comparison_betweenContext
    ):
        ctx.val = Comparison(
            lexpr=ctx.lhs.val,
            op=ComparisonOp.BETWEEN,
            rexpr=[ctx.low.val, ctx.high.val],
        )

    # Enter a parse tree produced by VerseQLParser#expression_comparison_in.
    def enterExpression_comparison_in(
        self, ctx: VerseQLParser.Expression_comparison_inContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#expression_comparison_in.
    def exitExpression_comparison_in(
        self, ctx: VerseQLParser.Expression_comparison_inContext
    ):
        op = ComparisonOp.IN if ctx.not_in is None else ComparisonOp.NIN
        args = ctx.operand()
        ctx.val = Comparison(
            lexpr=ctx.lhs.val,
            op=op,
            rexpr=[args[i].val for i in range(1, len(args))],
        )

    # Enter a parse tree produced by VerseQLParser#expression_not.
    def enterExpression_not(self, ctx: VerseQLParser.Expression_notContext):
        pass

    # Exit a parse tree produced by VerseQLParser#expression_not.
    def exitExpression_not(self, ctx: VerseQLParser.Expression_notContext):
        ctx.val = Not(expr=ctx.expression().val)

    # Enter a parse tree produced by VerseQLParser#expression_and.
    def enterExpression_and(self, ctx: VerseQLParser.Expression_andContext):
        pass

    # Exit a parse tree produced by VerseQLParser#expression_and.
    def exitExpression_and(self, ctx: VerseQLParser.Expression_andContext):
        ctx.val = And(lexpr=ctx.lhs.val, rexpr=ctx.rhs.val)

    # Enter a parse tree produced by VerseQLParser#expression_or.
    def enterExpression_or(self, ctx: VerseQLParser.Expression_orContext):
        pass

    # Exit a parse tree produced by VerseQLParser#expression_or.
    def exitExpression_or(self, ctx: VerseQLParser.Expression_orContext):
        ctx.val = Or(lexpr=ctx.lhs.val, rexpr=ctx.rhs.val)

        # Enter a parse tree produced by VerseQLParser#operand_value.

    def enterOperand_value(self, ctx: VerseQLParser.Operand_valueContext):
        pass

    # Exit a parse tree produced by VerseQLParser#operand_value.
    def exitOperand_value(self, ctx: VerseQLParser.Operand_valueContext):
        ctx.val = ctx.value().val

    # Enter a parse tree produced by VerseQLParser#operand_field.
    def enterOperand_field(self, ctx: VerseQLParser.Operand_fieldContext):
        pass

    # Exit a parse tree produced by VerseQLParser#operand_field.
    def exitOperand_field(self, ctx: VerseQLParser.Operand_fieldContext):
        ctx.val = ctx.field().val

    # Enter a parse tree produced by VerseQLParser#operand_parameter.
    def enterOperand_parameter(
        self, ctx: VerseQLParser.Operand_parameterContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#operand_parameter.
    def exitOperand_parameter(
        self, ctx: VerseQLParser.Operand_parameterContext
    ):
        ctx.val = ctx.parameter().val

    # Enter a parse tree produced by VerseQLParser#operand_ref.
    def enterOperand_ref(self, ctx: VerseQLParser.Operand_refContext):
        pass

    # Exit a parse tree produced by VerseQLParser#operand_ref.
    def exitOperand_ref(self, ctx: VerseQLParser.Operand_refContext):
        ctx.val = ctx.ref().val

    # Enter a parse tree produced by VerseQLParser#operand_function.
    def enterOperand_function(
        self, ctx: VerseQLParser.Operand_functionContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#operand_function.
    def exitOperand_function(self, ctx: VerseQLParser.Operand_functionContext):
        ctx.val = ctx.function().val

    # Enter a parse tree produced by VerseQLParser#order_by_clause.
    def enterOrder_by_clause(self, ctx: VerseQLParser.Order_by_clauseContext):
        pass

    # Exit a parse tree produced by VerseQLParser#order_by_clause.
    def exitOrder_by_clause(self, ctx: VerseQLParser.Order_by_clauseContext):
        ctx.key = "order_by"
        ctx.val = ctx.order_by().val

    # Enter a parse tree produced by VerseQLParser#rank_by_clause.
    def enterRank_by_clause(self, ctx: VerseQLParser.Rank_by_clauseContext):
        pass

    # Exit a parse tree produced by VerseQLParser#rank_by_clause.
    def exitRank_by_clause(self, ctx: VerseQLParser.Rank_by_clauseContext):
        ctx.key = "rank_by"
        ctx.val = ctx.expression().val

    # Enter a parse tree produced by VerseQLParser#order_by_terms.
    def enterOrder_by_terms(self, ctx: VerseQLParser.Order_by_termsContext):
        pass

    # Exit a parse tree produced by VerseQLParser#order_by_terms.
    def exitOrder_by_terms(self, ctx: VerseQLParser.Order_by_termsContext):
        ctx.val = OrderBy(terms=[term.val for term in ctx.order_by_term()])

    # Enter a parse tree produced by VerseQLParser#order_by_parameter.
    def enterOrder_by_parameter(
        self, ctx: VerseQLParser.Order_by_parameterContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#order_by_parameter.
    def exitOrder_by_parameter(
        self, ctx: VerseQLParser.Order_by_parameterContext
    ):
        ctx.val = ctx.parameter().val

    # Enter a parse tree produced by VerseQLParser#order_by_term.
    def enterOrder_by_term(self, ctx: VerseQLParser.Order_by_termContext):
        pass

    # Exit a parse tree produced by VerseQLParser#order_by_term.
    def exitOrder_by_term(self, ctx: VerseQLParser.Order_by_termContext):
        ctx.val = OrderByTerm(
            field=ctx.field().val.path,
            direction=(
                None if ctx.direction is None else ctx.direction.text.lower()
            ),
        )

    # Enter a parse tree produced by VerseQLParser#set_clause.
    def enterSet_clause(self, ctx: VerseQLParser.Set_clauseContext):
        pass

    # Exit a parse tree produced by VerseQLParser#set_clause.
    def exitSet_clause(self, ctx: VerseQLParser.Set_clauseContext):
        ctx.key = "set"
        ctx.val = ctx.update().val

    # Enter a parse tree produced by VerseQLParser#update_operations.
    def enterUpdate_operations(
        self, ctx: VerseQLParser.Update_operationsContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#update_operations.
    def exitUpdate_operations(
        self, ctx: VerseQLParser.Update_operationsContext
    ):
        ctx.val = Update(operations=[op.val for op in ctx.update_operation()])

    # Enter a parse tree produced by VerseQLParser#update_parameter.
    def enterUpdate_parameter(
        self, ctx: VerseQLParser.Update_parameterContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#update_parameter.
    def exitUpdate_parameter(self, ctx: VerseQLParser.Update_parameterContext):
        ctx.val = ctx.parameter().val

    # Enter a parse tree produced by VerseQLParser#update_operation.
    def enterUpdate_operation(
        self, ctx: VerseQLParser.Update_operationContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#update_operation.
    def exitUpdate_operation(self, ctx: VerseQLParser.Update_operationContext):
        args = ctx.function().val.args
        if args is None:
            ctx.val = UpdateOperation(
                field=ctx.field().val.path,
                op=ctx.function().val.name.lower(),
            )
        else:
            ctx.val = UpdateOperation(
                field=ctx.field().val.path,
                op=ctx.function().val.name.lower(),
                args=args,
            )

    # Enter a parse tree produced by VerseQLParser#function.
    def enterFunction(self, ctx: VerseQLParser.FunctionContext):
        pass

    # Exit a parse tree produced by VerseQLParser#function.
    def exitFunction(self, ctx: VerseQLParser.FunctionContext):
        args = ctx.function_args()
        if ctx.namespace is None:
            if isinstance(args.val, list):
                ctx.val = Function(name=ctx.name.text.lower(), args=args.val)
            elif isinstance(args.val, dict):
                ctx.val = Function(
                    name=ctx.name.text.lower(), named_args=args.val
                )
            else:
                ctx.val = Function(name=ctx.name.text.lower())
        else:
            if isinstance(args.val, list):
                ctx.val = Function(
                    name=ctx.name.text.lower(),
                    args=args.val,
                    namespace=ctx.namespace.text.lower(),
                )
            elif isinstance(args, dict):
                ctx.val = Function(
                    name=ctx.name.text.lower(),
                    named_args=args.val,
                    namespace=ctx.namespace.text.lower(),
                )
            else:
                ctx.val = Function(
                    ctx.name.text.lower(),
                    ctx.namespace.text.lower(),
                )

    # Enter a parse tree produced by VerseQLParser#function_no_args.
    def enterFunction_no_args(
        self, ctx: VerseQLParser.Function_no_argsContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#function_no_args.
    def exitFunction_no_args(self, ctx: VerseQLParser.Function_no_argsContext):
        ctx.val = None

    # Enter a parse tree produced by VerseQLParser#function_with_args.
    def enterFunction_with_args(
        self, ctx: VerseQLParser.Function_with_argsContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#function_with_args.
    def exitFunction_with_args(
        self, ctx: VerseQLParser.Function_with_argsContext
    ):
        ctx.val = [expr.val for expr in ctx.operand()]

    # Enter a parse tree produced by VerseQLParser#function_with_named_args.
    def enterFunction_with_named_args(
        self, ctx: VerseQLParser.Function_with_named_argsContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#function_with_named_args.
    def exitFunction_with_named_args(
        self, ctx: VerseQLParser.Function_with_named_argsContext
    ):
        kwargs = dict()
        for arg in ctx.named_arg():
            kwargs[arg.val[0]] = arg.val[1]
        ctx.val = kwargs

    # Enter a parse tree produced by VerseQLParser#named_arg.
    def enterNamed_arg(self, ctx: VerseQLParser.Named_argContext):
        pass

    # Exit a parse tree produced by VerseQLParser#named_arg.
    def exitNamed_arg(self, ctx: VerseQLParser.Named_argContext):
        ctx.val = (ctx.name.text, ctx.operand().val)

    # Enter a parse tree produced by VerseQLParser#ref.
    def enterRef(self, ctx: VerseQLParser.RefContext):
        pass

    # Exit a parse tree produced by VerseQLParser#ref.
    def exitRef(self, ctx: VerseQLParser.RefContext):
        ctx.val = Ref(path=ctx.path)

    # Enter a parse tree produced by VerseQLParser#parameter.
    def enterParameter(self, ctx: VerseQLParser.ParameterContext):
        pass

    # Exit a parse tree produced by VerseQLParser#parameter.
    def exitParameter(self, ctx: VerseQLParser.ParameterContext):
        ctx.val = Parameter(name=ctx.name.text)

    # Enter a parse tree produced by VerseQLParser#field.
    def enterField(self, ctx: VerseQLParser.FieldContext):
        pass

    # Exit a parse tree produced by VerseQLParser#field.
    def exitField(self, ctx: VerseQLParser.FieldContext):
        ctx.val = Field(path=ctx.getText())

    # Enter a parse tree produced by VerseQLParser#field_path.
    def enterField_path(self, ctx: VerseQLParser.Field_pathContext):
        pass

    # Exit a parse tree produced by VerseQLParser#field_path.
    def exitField_path(self, ctx: VerseQLParser.Field_pathContext):
        pass

    # Enter a parse tree produced by VerseQLParser#field_primitive.
    def enterField_primitive(self, ctx: VerseQLParser.Field_primitiveContext):
        pass

    # Exit a parse tree produced by VerseQLParser#field_primitive.
    def exitField_primitive(self, ctx: VerseQLParser.Field_primitiveContext):
        pass

    # Enter a parse tree produced by VerseQLParser#value_null.
    def enterValue_null(self, ctx: VerseQLParser.Value_nullContext):
        pass

    # Exit a parse tree produced by VerseQLParser#value_null.
    def exitValue_null(self, ctx: VerseQLParser.Value_nullContext):
        ctx.val = None

    # Enter a parse tree produced by VerseQLParser#value_true.
    def enterValue_true(self, ctx: VerseQLParser.Value_trueContext):
        pass

    # Exit a parse tree produced by VerseQLParser#value_true.
    def exitValue_true(self, ctx: VerseQLParser.Value_trueContext):
        ctx.val = True

    # Enter a parse tree produced by VerseQLParser#value_false.
    def enterValue_false(self, ctx: VerseQLParser.Value_falseContext):
        pass

    # Exit a parse tree produced by VerseQLParser#value_false.
    def exitValue_false(self, ctx: VerseQLParser.Value_falseContext):
        ctx.val = False

    # Enter a parse tree produced by VerseQLParser#value_string.
    def enterValue_string(self, ctx: VerseQLParser.Value_stringContext):
        pass

    # Exit a parse tree produced by VerseQLParser#value_string.
    def exitValue_string(self, ctx: VerseQLParser.Value_stringContext):
        ctx.val = eval(ctx.getText())

    # Enter a parse tree produced by VerseQLParser#value_integer.
    def enterValue_integer(self, ctx: VerseQLParser.Value_integerContext):
        pass

    # Exit a parse tree produced by VerseQLParser#value_integer.
    def exitValue_integer(self, ctx: VerseQLParser.Value_integerContext):
        ctx.val = int(ctx.getText())

    # Enter a parse tree produced by VerseQLParser#value_decimal.
    def enterValue_decimal(self, ctx: VerseQLParser.Value_decimalContext):
        pass

    # Exit a parse tree produced by VerseQLParser#value_decimal.
    def exitValue_decimal(self, ctx: VerseQLParser.Value_decimalContext):
        ctx.val = float(decimal.Decimal(ctx.getText()))

    # Enter a parse tree produced by VerseQLParser#value_json.
    def enterValue_json(self, ctx: VerseQLParser.Value_jsonContext):
        pass

    # Exit a parse tree produced by VerseQLParser#value_json.
    def exitValue_json(self, ctx: VerseQLParser.Value_jsonContext):
        ctx.val = json.loads(ctx.getText())

    # Enter a parse tree produced by VerseQLParser#value_array.
    def enterValue_array(self, ctx: VerseQLParser.Value_arrayContext):
        pass

    # Exit a parse tree produced by VerseQLParser#value_array.
    def exitValue_array(self, ctx: VerseQLParser.Value_arrayContext):
        ctx.val = ctx.array().val

    # Enter a parse tree produced by VerseQLParser#array_empty.
    def enterArray_empty(self, ctx: VerseQLParser.Array_emptyContext):
        pass

    # Exit a parse tree produced by VerseQLParser#array_empty.
    def exitArray_empty(self, ctx: VerseQLParser.Array_emptyContext):
        ctx.val = []

    # Enter a parse tree produced by VerseQLParser#array_items.
    def enterArray_items(self, ctx: VerseQLParser.Array_itemsContext):
        pass

    # Exit a parse tree produced by VerseQLParser#array_items.
    def exitArray_items(self, ctx: VerseQLParser.Array_itemsContext):
        ctx.val = [item.val for item in ctx.value()]

    # Enter a parse tree produced by VerseQLParser#json.
    def enterJson(self, ctx: VerseQLParser.JsonContext):
        pass

    # Exit a parse tree produced by VerseQLParser#json.
    def exitJson(self, ctx: VerseQLParser.JsonContext):
        pass

    # Enter a parse tree produced by VerseQLParser#json_obj.
    def enterJson_obj(self, ctx: VerseQLParser.Json_objContext):
        pass

    # Exit a parse tree produced by VerseQLParser#json_obj.
    def exitJson_obj(self, ctx: VerseQLParser.Json_objContext):
        pass

    # Enter a parse tree produced by VerseQLParser#json_pair.
    def enterJson_pair(self, ctx: VerseQLParser.Json_pairContext):
        pass

    # Exit a parse tree produced by VerseQLParser#json_pair.
    def exitJson_pair(self, ctx: VerseQLParser.Json_pairContext):
        pass

    # Enter a parse tree produced by VerseQLParser#json_arr.
    def enterJson_arr(self, ctx: VerseQLParser.Json_arrContext):
        pass

    # Exit a parse tree produced by VerseQLParser#json_arr.
    def exitJson_arr(self, ctx: VerseQLParser.Json_arrContext):
        pass

    # Enter a parse tree produced by VerseQLParser#json_value.
    def enterJson_value(self, ctx: VerseQLParser.Json_valueContext):
        pass

    # Exit a parse tree produced by VerseQLParser#json_value.
    def exitJson_value(self, ctx: VerseQLParser.Json_valueContext):
        pass


del VerseQLParser



from __future__ import annotations

import json
from typing import Any

from .data_model import DataModel


class Operation(DataModel):
    """Operation.

    Attributes:
        name: Operation name.
        args: Operation arguments.
    """

    name: str | None = None
    args: dict[str, Any] | None = None

    @staticmethod
    def normalize(
        name: str | None,
        args: dict[str, Any] | None,
    ) -> Operation:
        if args is None:
            return Operation(name=name)
        if "self" in args:
            args.pop("self")
        rargs: dict = {}
        for k, v in args.items():
            if k == "kwargs":
                rargs.update(v)
            elif v is not None:
                rargs[k] = v
        return Operation(name=name, args=rargs)

    @staticmethod
    def create_with_func_name(args: dict[str, Any] | None) -> Operation:
        import inspect

        name = None
        frame = inspect.currentframe()
        if frame is not None:
            if frame.f_back is not None:
                name = frame.f_back.f_code.co_name

        return Operation.normalize(name=name, args=args)

    def __str__(self) -> str:
        if self.name:
            str = self.name
        else:
            str = ""
        if self.args is not None:
            for key, value in self.args.items():
                nkey = key.replace("_", " ").strip()
                if isinstance(value, list) and all(
                    isinstance(a, Operation) for a in value
                ):
                    nvalue = "; ".join([self._str_value(a) for a in value])
                else:
                    nvalue = self._str_value(value)
                str = f"{str} {nkey} {nvalue}"
        return str

    def _str_value(self, value: Any) -> Any:
        if value is None or isinstance(
            value, (str, int, float, bool, dict, list)
        ):
            return json.dumps(value)
        return str(value)



import copy
import re
from typing import Any, Callable

from ..core._data_accessor import DataAccessor
from ..core.exceptions import BadRequestError
from ._functions import QueryFunctionName
from ._models import (
    And,
    Comparison,
    ComparisonOp,
    Expression,
    Field,
    Function,
    FunctionNamespace,
    Not,
    Or,
    OrderBy,
    OrderByDirection,
    Select,
    Undefined,
    Update,
    UpdateOp,
    UpdateOperation,
    Value,
)
from ._ql_parser import QLParser


class QueryProcessor:
    @staticmethod
    def count_items(
        items: list[Any],
        where: str | Expression | None = None,
        field_resolver: Callable | None = None,
    ) -> int:
        current_items = items
        current_items = QueryProcessor.filter_items(
            current_items, where, field_resolver
        )
        return len(current_items)

    @staticmethod
    def query_items(
        items: list[Any],
        select: str | Select | None = None,
        where: str | Expression | None = None,
        order_by: str | OrderBy | None = None,
        limit: int | None = None,
        offset: int | None = None,
        field_resolver: Callable | None = None,
    ) -> list[Any]:
        current_items = items
        current_items = QueryProcessor.filter_items(
            current_items, where, field_resolver
        )
        current_items = QueryProcessor.order_items(
            current_items, order_by, field_resolver
        )
        current_items = QueryProcessor.project_items(
            current_items, select, field_resolver
        )
        current_items = QueryProcessor.limit_items(
            current_items, limit, offset
        )
        return current_items

    @staticmethod
    def filter_items(
        items: list[Any],
        where: str | Expression | None = None,
        field_resolver: Callable | None = None,
    ) -> list[Any]:
        result = []
        for item in items:
            if QueryProcessor.filter_item(item, where, field_resolver):
                result.append(item)
        return result

    @staticmethod
    def filter_item(
        item: Any,
        where: str | Expression | None = None,
        field_resolver: Callable | None = None,
    ) -> Any:
        if where is None:
            return item
        if not isinstance(where, str):
            expr: Expression = where
        else:
            expr = QLParser.parse_where(where)
        if bool(QueryProcessor.eval_expr(item, expr, field_resolver)):
            return item
        return None

    @staticmethod
    def order_items(
        items: list[Any],
        order_by: str | OrderBy | None = None,
        field_resolver: Callable | None = None,
    ) -> list[Any]:
        if order_by is None:
            return items
        if isinstance(order_by, OrderBy):
            ob: OrderBy | None = order_by
        else:
            ob = QLParser.parse_order_by(order_by)
        if ob is None:
            return items
        filtered_items = items
        for term in ob.terms:
            if isinstance(term.field, str):
                filtered_items = QueryProcessor.filter_items(
                    filtered_items,
                    f"is_defined({term.field})",
                    field_resolver,
                )

        def comparer(item):
            values = []
            for term in order_by.terms:
                value = QueryProcessor.eval_expr(
                    item, Field(path=term.field), field_resolver
                )
                if term.direction == OrderByDirection.DESC:
                    values.append(DescFieldComparer(value))
                else:
                    values.append(AscFieldComparer(value))
            return tuple(values)

        sorted_items = sorted(filtered_items, key=comparer)
        return sorted_items

    @staticmethod
    def limit_items(
        items: list[Any],
        limit: int | None,
        offset: int | None,
    ) -> list[Any]:
        start = 0
        end = len(items)
        if offset is not None:
            start = offset
        if limit is not None:
            end = start + limit
        return items[start:end]

    @staticmethod
    def project_items(
        items: list[Any],
        select: str | Select | None,
        field_resolver: Callable | None = None,
    ) -> list[Any]:
        result = []
        for item in items:
            result.append(
                QueryProcessor.project_item(item, select, field_resolver)
            )
        return result

    @staticmethod
    def project_item(
        item: Any,
        select: str | Select | None,
        field_resolver: Callable | None = None,
    ) -> Any:
        result: dict = dict()
        if select is None:
            return item
        if isinstance(select, Select):
            sel: Select | None = select
        else:
            sel = QLParser.parse_select(select)
        if sel is None or sel.terms is None or len(sel.terms) == 0:
            return item
        for term in sel.terms:
            val = QueryProcessor.eval_field(
                item, Field(path=term.field), field_resolver
            )
            if isinstance(val, Undefined):
                continue
            alias = term.alias if term.alias is not None else term.field
            QueryProcessor.update_field(
                result,
                UpdateOperation(field=alias, op=UpdateOp.PUT, args=[val]),
                field_resolver,
            )
        return result

    @staticmethod
    def eval_field(
        item: Any,
        field: Field,
        field_resolver: Callable | None = None,
    ) -> Value | Undefined:
        path = field.path
        if field_resolver is not None:
            path = field_resolver(field.path)
        return DataAccessor.get_field(item, path)

    @staticmethod
    def eval_function(
        item: Any,
        expr: Function,
        field_resolver: Callable | None = None,
    ) -> Value:
        if expr.namespace != FunctionNamespace.BUILTIN:
            raise BadRequestError("Function not supported")
        if expr.name == QueryFunctionName.EXISTS:
            return item is not None
        if expr.name == QueryFunctionName.NOT_EXISTS:
            return item is None
        if expr.name == QueryFunctionName.IS_TYPE:
            val = QueryProcessor.eval_expr(item, expr.args[0], field_resolver)
            type = expr.args[1]
            if not isinstance(type, str):
                raise BadRequestError(
                    "type in IS_TYPE function must be specified as a string"
                )
            if type == "string":
                return isinstance(val, str)
            if type == "number":
                return isinstance(val, (int, float))
            if type == "boolean":
                return isinstance(val, bool)
            if type == "array":
                return isinstance(val, list)
            if type == "object":
                return isinstance(val, dict)
            if type == "null":
                return val is None
            raise BadRequestError("Unknown type passed to IS_TYPE function")
        if expr.name == QueryFunctionName.IS_DEFINED:
            val = QueryProcessor.eval_expr(item, expr.args[0], field_resolver)
            if isinstance(val, Undefined):
                return False
            return True
        if expr.name == QueryFunctionName.IS_NOT_DEFINED:
            val = QueryProcessor.eval_expr(item, expr.args[0], field_resolver)
            if isinstance(val, Undefined):
                return True
            return False
        if expr.name == QueryFunctionName.LENGTH:
            val = QueryProcessor.eval_expr(item, expr.args[0], field_resolver)
            if val is None:
                return 0
            if isinstance(val, str):
                return len(val)
            return 0
        if expr.name == QueryFunctionName.CONTAINS:
            val1 = QueryProcessor.eval_expr(item, expr.args[0], field_resolver)
            val2 = QueryProcessor.eval_expr(item, expr.args[1], field_resolver)
            if val1 is None:
                return False
            if isinstance(val1, str) and isinstance(val2, str):
                return val2 in val1
            return False
        if expr.name == QueryFunctionName.STARTS_WITH:
            val1 = QueryProcessor.eval_expr(item, expr.args[0], field_resolver)
            val2 = QueryProcessor.eval_expr(item, expr.args[1], field_resolver)
            if val1 is None:
                return False
            if isinstance(val1, str) and isinstance(val2, str):
                return val1.startswith(val2)
            return False
        if expr.name == QueryFunctionName.ARRAY_LENGTH:
            val = QueryProcessor.eval_expr(item, expr.args[0], field_resolver)
            if val is None:
                return 0
            if isinstance(val, list):
                return len(val)
            return 0
        if expr.name == QueryFunctionName.ARRAY_CONTAINS:
            val1 = QueryProcessor.eval_expr(item, expr.args[0], field_resolver)
            val2 = QueryProcessor.eval_expr(item, expr.args[1], field_resolver)
            if val1 is None:
                return False
            if isinstance(val1, list):
                return val2 in val1
            return False
        if expr.name == QueryFunctionName.ARRAY_CONTAINS_ANY:
            val1 = QueryProcessor.eval_expr(item, expr.args[0], field_resolver)
            val2 = QueryProcessor.eval_expr(item, expr.args[1], field_resolver)
            if val1 is None:
                return False
            if isinstance(val1, list) and isinstance(val2, list):
                for item in val2:
                    if item in val1:
                        return True
            return False
        if expr.name == QueryFunctionName.RANDOM:
            import random

            return random.random()
        if expr.name == QueryFunctionName.NOW:
            from datetime import datetime, timezone

            now = datetime.now(timezone.utc)
            formatted_now = now.strftime("%Y-%m-%d %H:%M:%S.%f%z")

            return formatted_now
        raise BadRequestError(f"Function {expr.name} not supported")

    @staticmethod
    def eval_comparison(
        item: Any | None,
        expr: Comparison,
        field_resolver: Callable | None = None,
    ) -> bool:
        lval = QueryProcessor.eval_expr(item, expr.lexpr, field_resolver)
        rval = QueryProcessor.eval_expr(item, expr.rexpr, field_resolver)
        if (
            expr.op == ComparisonOp.LT
            or expr.op == ComparisonOp.LTE
            or expr.op == ComparisonOp.GT
            or expr.op == ComparisonOp.GTE
        ):
            if isinstance(lval, str) and isinstance(rval, str):
                if expr.op == ComparisonOp.LT:
                    return lval < rval
                if expr.op == ComparisonOp.LTE:
                    return lval <= rval
                if expr.op == ComparisonOp.GT:
                    return lval > rval
                if expr.op == ComparisonOp.GTE:
                    return lval >= rval
            if isinstance(lval, (int, float)) and isinstance(
                rval, (int, float)
            ):
                if expr.op == ComparisonOp.LT:
                    return lval < rval
                if expr.op == ComparisonOp.LTE:
                    return lval <= rval
                if expr.op == ComparisonOp.GT:
                    return lval > rval
                if expr.op == ComparisonOp.GTE:
                    return lval >= rval
            return False
        elif expr.op == ComparisonOp.EQ:
            return lval == rval
        elif expr.op == ComparisonOp.NEQ:
            return lval != rval
        elif expr.op == ComparisonOp.BETWEEN:
            if isinstance(rval, list):
                rval1 = rval[0]
                rval2 = rval[1]
                if (
                    isinstance(lval, str)
                    and isinstance(rval1, str)
                    and isinstance(rval2, str)
                ):
                    return lval >= rval1 and lval <= rval2
                if (
                    isinstance(lval, (int, float))
                    and isinstance(rval1, (int, float))
                    and isinstance(rval2, (int, float))
                ):
                    return lval >= rval1 and lval <= rval2
                return False
            raise BadRequestError(
                """Right operand for BETWEEN must be a list with two values"""
            )
        elif expr.op == ComparisonOp.IN:
            if isinstance(rval, list):
                return lval in rval
            raise BadRequestError(
                "Right operand for IN comparison must be a list"
            )
        elif expr.op == ComparisonOp.NIN:
            if isinstance(rval, list):
                return lval not in rval
            raise BadRequestError(
                "Right operand for NOT IN comparison must be a list"
            )
        elif expr.op == ComparisonOp.LIKE:
            if isinstance(lval, str) and isinstance(rval, str):
                return bool(re.match(rval, lval))
            return False
        raise BadRequestError("Comparison not supported")

    @staticmethod
    def eval_expr(
        item: Any,
        expr: Expression,
        field_resolver: Callable | None = None,
    ) -> Value | Undefined:
        if expr is None:
            return True
        if isinstance(expr, (bool, int, float, str, list, dict)):
            return expr
        if isinstance(expr, Field):
            return QueryProcessor.eval_field(item, expr, field_resolver)
        if isinstance(expr, Function):
            return QueryProcessor.eval_function(item, expr, field_resolver)
        if isinstance(expr, Comparison):
            return QueryProcessor.eval_comparison(item, expr, field_resolver)
        if isinstance(expr, And):
            return bool(
                QueryProcessor.eval_expr(item, expr.lexpr, field_resolver)
            ) and bool(
                QueryProcessor.eval_expr(item, expr.rexpr, field_resolver)
            )
        if isinstance(expr, Or):
            return bool(
                QueryProcessor.eval_expr(item, expr.lexpr, field_resolver)
            ) or bool(
                QueryProcessor.eval_expr(item, expr.rexpr, field_resolver)
            )
        if isinstance(expr, Not):
            return not bool(
                QueryProcessor.eval_expr(item, expr.expr, field_resolver)
            )
        raise BadRequestError(f"Expression {str(expr)} not supported")

    @staticmethod
    def update_item(
        item: Any,
        update: str | Update,
        field_resolver: Callable | None = None,
    ) -> dict:
        if isinstance(update, Update):
            up: Update | None = update
        else:
            up = QLParser.parse_update(update)
        if up is None:
            return item
        item_copy = copy.deepcopy(item)
        for operation in up.operations:
            QueryProcessor.update_field(item_copy, operation, field_resolver)
        return item_copy

    @staticmethod
    def update_field(
        value: Any,
        operation: UpdateOperation,
        field_resolver: Callable | None = None,
    ) -> None:
        field = operation.field
        op = operation.op
        path = field
        if field_resolver is not None:
            path = field_resolver(path)
        DataAccessor.update_field(
            value,
            path,
            op,
            (
                operation.args[0]
                if operation.args is not None and len(operation.args) > 0
                else None
            ),
        )

    @staticmethod
    def extract_filter_fields(
        where: Expression, field_resolver: Callable | None = None
    ):
        fields = []

        def extract_field(expr: Expression):
            if isinstance(expr, Field):
                field = (
                    expr.path
                    if field_resolver is None
                    else field_resolver(expr.path)
                )
                field = field.split("[")[0]
                if field not in fields:
                    fields.append(field)
            if isinstance(expr, Function):
                for arg in expr.args:
                    extract_field(arg)
            if isinstance(expr, Comparison):
                extract_field(expr.lexpr)
                extract_field(expr.rexpr)
            if isinstance(expr, And):
                extract_field(expr.lexpr)
                extract_field(expr.rexpr)
            if isinstance(expr, Or):
                extract_field(expr.lexpr)
                extract_field(expr.rexpr)
            if isinstance(expr, Not):
                extract_field(expr.expr)
            return

        extract_field(where)
        return fields


class AscFieldComparer:
    def __init__(self, obj):
        self.obj = obj

    def __eq__(self, other):
        return self.obj == other.obj

    def __lt__(self, other):
        return self.obj < other.obj


class DescFieldComparer:
    def __init__(self, obj):
        self.obj = obj

    def __eq__(self, other):
        return self.obj == other.obj

    def __lt__(self, other):
        return self.obj > other.obj


def default_field_resolver(field: str) -> str:
    return field



from __future__ import annotations

from typing import Any

from .data_model import DataModel
from .manifest import MANIFEST_FILE


class Context(DataModel):
    id: str | None = None
    data: dict[str, Any] | None = None


class RunContext(DataModel):
    """
    Run context.
    """

    handle: str | None = None
    tag: str | None = None
    path: str = "."
    manifest: str = MANIFEST_FILE
    component_type: str | None = None



__all__ = ["Time"]


import time


class Time:
    @staticmethod
    def now() -> float:
        timestamp = time.time()
        return timestamp

    @staticmethod
    def now_ms() -> float:
        timestamp_ms = time.time_ns() / 1_000_000
        return timestamp_ms



from __future__ import annotations

from .data_model import DataModel


class ComponentNode(DataModel):
    handle: str
    type: str
    kind: str
    requirements: list[str]
    provider: ProviderNode | None = None
    depends: list[ComponentEdge] = []


class ProviderNode(DataModel):
    handle: str
    type: str
    kind: str
    requirements: list[str]
    composes: list[ProviderEdge] = []
    depends: list[ComponentEdge] = []


class ComponentEdge(DataModel):
    component: ComponentNode
    ref: str


class ProviderEdge(DataModel):
    provider: ProviderNode
    ref: str



import argparse
import sys
from typing import Any

from ._arg_parser import ArgParser
from ._async_helper import run_sync
from ._context import Context, RunContext
from ._loader import Loader
from .manifest import MANIFEST_FILE
from .spec import SpecBuilder


def run(
    path: str,
    manifest: str,
    handle: str,
    tag: str | None,
    statement: str | None,
) -> Any:
    """
    Verse Run
    """
    loader = Loader(path=path, manifest=manifest)
    component_type = loader.get_component_type(handle=handle)
    component = loader.load_component(handle=handle, tag=tag)
    run_context = RunContext(
        handle=handle,
        tag=tag,
        path=path,
        manifest=manifest,
        component_type=component_type,
    )
    operation = None
    if statement is not None:
        operation = ArgParser.convert_execute_operation(statement, None)
    return component.__run__(
        operation=operation,
        context=Context(data=dict(__run__=run_context)),
    )


async def arun(
    path: str,
    manifest: str,
    handle: str,
    tag: str | None,
    statement: str | None,
) -> Any:
    """
    Verse Run Async
    """
    loader = Loader(path=path, manifest=manifest)
    component_type = loader.get_component_type(handle=handle)
    component = loader.load_component(handle=handle, tag=tag)
    run_context = RunContext(
        handle=handle,
        tag=tag,
        path=path,
        manifest=manifest,
        component_type=component_type,
    )
    operation = None
    if statement is not None:
        operation = ArgParser.convert_execute_operation(statement, None)
    return await component.__arun__(
        operation=operation,
        context=Context(data=dict(__run__=run_context)),
    )


def requirements(
    path: str,
    manifest: str,
    handle: str,
    tag: str | None,
    out: str | None,
):
    """
    Verse Requirements
    """
    loader = Loader(path=path, manifest=manifest)
    requirements = loader.generate_requirements(
        handle=handle, tag=tag, out=out
    )
    return requirements


def spec(
    path: str,
    manifest: str,
    type_or_handle: str,
):
    """
    Verse Spec
    """
    spec_builder = SpecBuilder(path=path)
    if "." in type_or_handle:
        component_type = type_or_handle
    else:
        component_type = Loader(
            path=path, manifest=manifest
        ).get_component_type(handle=type_or_handle)
    component_spec = spec_builder.build_component_spec(component_type)
    spec_builder.print(component_spec)


def main():
    parser = argparse.ArgumentParser(prog="verse", description="Verse CLI")
    subparsers = parser.add_subparsers(dest="command", required=True)
    run_parser = subparsers.add_parser("run", help="Run the Verse application")
    arun_parser = subparsers.add_parser(
        "arun", help="Run the Verse application in async mode"
    )
    requirements_parser = subparsers.add_parser(
        "requirements", help="Generate the pip requirements"
    )
    spec_parser = subparsers.add_parser("spec", help="Get component spec")
    run_parser_arguments = [
        (
            "handle",
            str,
            None,
            "Component handle (required positional argument)",
            None,
        ),
        (
            "tag",
            str,
            None,
            "Binding tag (optional positional argument)",
            "?",
        ),
        ("--path", str, ".", "Project directory", None),
        ("--manifest", str, MANIFEST_FILE, "Manifest filename", None),
        ("--execute", str, None, "Operation to execute", None),
    ]
    requirements_parser_arguments = [
        (
            "handle",
            str,
            None,
            "Component handle (required positional argument)",
            None,
        ),
        (
            "tag",
            str,
            None,
            "Binding tag (optional positional argument)",
            "?",
        ),
        ("--path", str, ".", "Project directory", None),
        ("--manifest", str, MANIFEST_FILE, "Manifest filename", None),
        ("--out", str, None, "Output path", None),
    ]
    spec_parser_arguments = [
        ("type_or_handle", str, None, "Component type or handle", "?"),
        ("--path", str, ".", "Project directory", None),
        ("--manifest", str, MANIFEST_FILE, "Manifest filename", None),
    ]
    for arg in run_parser_arguments:
        run_parser.add_argument(
            arg[0], type=arg[1], default=arg[2], help=arg[3], nargs=arg[4]
        )
        arun_parser.add_argument(
            arg[0], type=arg[1], default=arg[2], help=arg[3], nargs=arg[4]
        )
    for arg in requirements_parser_arguments:
        requirements_parser.add_argument(
            arg[0], type=arg[1], default=arg[2], help=arg[3], nargs=arg[4]
        )
    for arg in spec_parser_arguments:
        spec_parser.add_argument(
            arg[0], type=arg[1], default=arg[2], help=arg[3], nargs=arg[4]
        )

    args = parser.parse_args()
    if args.command == "run":
        response = run(
            path=args.path,
            manifest=args.manifest,
            handle=args.handle,
            tag=args.tag,
            statement=args.execute,
        )
        if response:
            print(response)
    elif args.command == "arun":
        try:
            response = run_sync(
                arun,
                path=args.path,
                manifest=args.manifest,
                handle=args.handle,
                tag=args.tag,
                statement=args.execute,
            )
            if response:
                print(response)
        except KeyboardInterrupt:
            sys.exit(0)
    elif args.command == "requirements":
        requirements(
            path=args.path,
            manifest=args.manifest,
            handle=args.handle,
            tag=args.tag,
            out=args.out,
        )
    elif args.command == "spec":
        spec(
            path=args.path,
            manifest=args.manifest,
            type_or_handle=args.type_or_handle,
        )
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



from __future__ import annotations

import ast
import importlib
import inspect
import os
import sys
from enum import Enum
from typing import Any, Literal, get_type_hints

from ._component import Component, Provider
from ._loader import Loader
from ._yaml_loader import YamlLoader
from .data_model import DataModel, Empty, get_args, get_origin
from .exceptions import NotFoundError

__all__ = [
    "Kind",
    "ComponentSpec",
    "ContentModal",
    "OperationArgSpec",
    "OperationExceptionSpec",
    "OperationReturnSpec",
    "OperationSpec",
    "ParameterOptionSpec",
    "ParameterSpec",
    "ProviderMode",
    "ProviderSpec",
    "PyInfo",
    "TypeAttributeSpec",
    "TypeSpec",
]


class Kind(str, Enum):
    BASE = "base"
    CUSTOM = "custom"
    COMPOSITE = "composite"


class ContentModal(str, Enum):
    TEXT = "text"
    IMAGE = "image"
    VIDEO = "video"
    AUDIO = "audio"
    MULTI = "multi"


class ProviderMode(str, Enum):
    MIXED = "mixed"
    LOCAL = "local"
    SERVICE = "service"


class TypeInfo(DataModel):
    name: str
    args: list[str | TypeInfo] | None = None

    def __repr__(self):
        if self.args:
            if self.name == "Union":
                return " | ".join(repr(arg) for arg in self.args)
            formatted_args = ", ".join(repr(arg) for arg in self.args)
            return f"{self.name}[{formatted_args}]"
        return self.name

    def __str__(self):
        return self.__repr__()


class ParameterOptionSpec(DataModel):
    label: str
    value: Any


class ParameterSpec(DataModel):
    name: str
    type: TypeInfo | None = None
    description: str | None = None
    required: bool = True
    default: Any = None
    options: list[ParameterOptionSpec] = []


class OperationArgSpec(DataModel):
    name: str
    type: TypeInfo | None = None
    description: str | None = None
    required: bool = True
    default: Any = None


class OperationReturnSpec(DataModel):
    type: TypeInfo | None = None
    description: str | None = None


class OperationExceptionSpec(DataModel):
    type: TypeInfo | None = None
    description: str | None = None


class OperationSpec(DataModel):
    name: str
    args: list[OperationArgSpec] = []
    description: str | None = None
    ret: OperationReturnSpec | None = None
    exceptions: list[OperationExceptionSpec] = []


class TypeAttributeSpec(DataModel):
    name: str
    type: TypeInfo | None = None
    description: str | None = None
    required: bool | None = True
    default: Any = None


class TypeSpec(DataModel):
    name: str
    description: str | None = None
    attributes: list[TypeAttributeSpec] = []
    inherits: list[TypeSpec] = []


class PyInfo(DataModel):
    package: str | None = None
    path: str | None = None
    requirements: list[str] = []


class ProviderSpec(DataModel):
    type: str
    version: str | None = None
    title: str | None = None
    description: str | None = None
    logo: str | None = None
    author: str | None = None
    readme: str | None = None
    link: str | None = None
    kind: Kind = Kind.BASE
    mode: ProviderMode = ProviderMode.MIXED
    parameters: list[ParameterSpec] = []
    py: PyInfo | None = None


class ComponentSpec(DataModel):
    type: str
    version: str | None = None
    title: str | None = None
    description: str | None = None
    logo: str | None = None
    author: str | None = None
    readme: str | None = None
    link: str | None = None
    category: str | None = None
    modals: list[ContentModal] = []
    kind: Kind = Kind.BASE
    parameters: list[ParameterSpec] = []
    operations: list[OperationSpec] = []
    providers: list[ProviderSpec] = []
    types: list[TypeSpec] = []
    py: PyInfo | None = None

    @staticmethod
    def parse(path: str) -> ComponentSpec:
        obj = YamlLoader.load(path=path)
        manifest = ComponentSpec.from_dict(obj)
        return manifest

    def get_provider(self, name: str) -> ProviderSpec:
        for provider in self.providers:
            if name == provider.type:
                return provider
        raise NotFoundError("Provider not found")


class SpecBuilder:
    def __init__(
        self,
        path: str = ".",
    ):
        abs_project_folder = os.path.abspath(path)
        if abs_project_folder not in sys.path:
            sys.path.append(abs_project_folder)

    def build_component_spec(
        self,
        component_type: str,
    ) -> ComponentSpec:
        py_path = Loader.get_component_path(component_type)
        kind = Kind(Loader.get_component_kind(component_type))
        component = Loader.load_class(py_path, Component)
        title = component_type
        description = SpecBuilder._get_top_level_docstring_from_class(
            component
        )
        types = SpecBuilder._get_operations_custom_types(component)
        types = SpecBuilder._combine_types_list(
            types, SpecBuilder._get_parameters_custom_types(component)
        )
        types = SpecBuilder._add_dependent_types(types)
        type_specs = [SpecBuilder._get_type(type) for type in types]

        provider_types = SpecBuilder._get_provider_types(py_path)
        providers = []
        for provider_type in provider_types:
            try:
                providers.append(
                    self.build_provider_spec(component_type, provider_type)
                )
                types = SpecBuilder._combine_types_list(
                    types,
                    SpecBuilder._get_provider_parameters_custom_types(
                        component_type, provider_type
                    ),
                )
            except Exception as e:
                print(e)
        spec = ComponentSpec(
            type=component_type,
            title=title,
            description=description,
            kind=kind,
            category=None,
            operations=SpecBuilder._get_operations(component),
            parameters=SpecBuilder._get_parameters(component),
            providers=providers,
            types=type_specs,
            py=PyInfo(path=py_path),
        )
        return spec

    def build_provider_spec(
        self,
        component_type: str,
        provider_type: str,
    ) -> ProviderSpec:
        py_path = Loader.get_provider_path(component_type, provider_type)
        kind = Loader.get_provider_kind(component_type, provider_type)
        provider = Loader.load_class(py_path, Provider)
        title = provider_type
        description = SpecBuilder._get_top_level_docstring_from_class(provider)
        spec = ProviderSpec(
            type=provider_type,
            title=title,
            description=description,
            kind=kind,
            parameters=SpecBuilder._get_parameters(provider),
            py=PyInfo(path=py_path),
        )
        return spec

    def print(self, component_spec: ComponentSpec):
        print()
        title = component_spec.description or component_spec.title
        print(f"{title}")
        if title:
            print("-" * len(title))
        SpecBuilder._print_parameters(component_spec.parameters)
        SpecBuilder._print_operations(component_spec.operations)
        SpecBuilder._print_providers(component_spec.providers)
        SpecBuilder._print_types(component_spec.types)

    @staticmethod
    def _print_parameters(specs: list[ParameterSpec]):
        parameters = []
        for param in specs:
            if param.description:
                description = param.description.strip().replace("\n", " ")
            else:
                description = ""
            type = str(param.type)
            if len(param.options) > 0:
                options_str = ", ".join(
                    [option.value for option in param.options]
                )
                type = f"{type} ({options_str})"
            if param.required:
                type = f"* {type}"
            if not param.required:
                type = f"{type} ={str(param.default)}"
            parameters.append((param.name, description, type))

        if parameters:
            col_widths = [
                max(max(len(row[0]) for row in parameters), 12),
                max(max(len(row[1]) for row in parameters), 12),
                max(max(len(row[2]) for row in parameters), 12),
            ]
            print(
                f"{'Parameter'.ljust(col_widths[0])}  "
                f"{'Description'.ljust(col_widths[1])}  "
                f"{'Type'}"
            )
            print(
                f"{'---------'.ljust(col_widths[0])}  "
                f"{'-----------'.ljust(col_widths[1])}  "
                f"{'-----'}"
            )
            for name, desc, type in parameters:
                print(
                    f"{name.ljust(col_widths[0])}  "
                    f"{desc.ljust(col_widths[1])}  "
                    f"{type}"
                )
            print()

    @staticmethod
    def _print_operations(specs: list[OperationSpec]):
        operations = []
        for op in specs:
            if op.description:
                description = op.description.strip().replace("\n", " ")
            else:
                description = ""
            usage = op.name
            for arg in op.args:
                arg_usage = f"{arg.name} <{arg.type}>"
                if not arg.required:
                    arg_usage = f"[{arg_usage}]"
                usage = f"{usage} {arg_usage}"
            operations.append((op.name, description, usage))

        if operations:
            col_widths = [
                max(max(len(row[0]) for row in operations), 12),
                max(max(len(row[1]) for row in operations), 12),
                max(max(len(row[2]) for row in operations), 12),
            ]
            print(
                f"{'Operation'.ljust(col_widths[0])}  "
                f"{'Description'.ljust(col_widths[1])}  "
                f"{'Usage'}"
            )
            print(
                f"{'---------'.ljust(col_widths[0])}  "
                f"{'-----------'.ljust(col_widths[1])}  "
                f"{'-----'}"
            )

            for name, desc, usage in operations:
                usage_lines = usage.split("\n")
                print(
                    f"{name.ljust(col_widths[0])}  "
                    f"{desc.ljust(col_widths[1])}  "
                    f"{usage_lines[0]}"
                )
                for extra_line in usage_lines[1:]:
                    print(
                        " " * (col_widths[0] + 2 + col_widths[1] + 2)
                        + extra_line
                    )
        print()

    @staticmethod
    def _print_providers(specs: list[ProviderSpec]):
        print("Providers")
        print("---------")
        for provider in specs:
            print(provider.type)
            if provider.description:
                print(provider.description)
            print("------")
            SpecBuilder._print_parameters(provider.parameters)

    @staticmethod
    def _print_types(specs: list[TypeSpec]):
        for spec in specs:
            SpecBuilder._print_type(spec)

    @staticmethod
    def _print_type(spec: TypeSpec):
        if spec.description:
            desc = spec.description.strip().replace("\n", " ")
        else:
            desc = ""
        title = f"{spec.name}"
        if spec.inherits:
            inherits = ", ".join([f"{base.name}" for base in spec.inherits])
            title = f"{title} ({inherits})"
        print(title)
        print("-" * len(title))

        attributes = []
        for attr in spec.attributes:
            if attr.description:
                description = attr.description.strip().replace("\n", " ")
            else:
                description = ""
            type = str(attr.type)
            if attr.required:
                type = f"* {type}"
            if not attr.required:
                type = f"{type} ={str(attr.default)}"
            attributes.append((attr.name, description, type))

        if attributes:
            col_widths = [
                max(max(len(row[0]) for row in attributes), 12),
                max(max(len(row[1]) for row in attributes), 12),
                max(max(len(row[2]) for row in attributes), 12),
            ]
            print(
                f"{'Attribute'.ljust(col_widths[0])}  "
                f"{'Description'.ljust(col_widths[1])}  "
                f"{'Type'}"
            )
            print(
                f"{'---------'.ljust(col_widths[0])}  "
                f"{'-----------'.ljust(col_widths[1])}  "
                f"{'-----'}"
            )
            for name, desc, type in attributes:
                print(
                    f"{name.ljust(col_widths[0])}  "
                    f"{desc.ljust(col_widths[1])}  "
                    f"{type}"
                )
        print()

    @staticmethod
    def _get_provider_types(component_package: str) -> list[str]:
        component_path = SpecBuilder._get_component_file_path(
            component_package
        )
        if ":" in component_package:
            return []
        providers_path = os.path.join(component_path, "providers")
        providers = [
            f[:-3]
            for f in os.listdir(providers_path)
            if f.endswith(".py")
            and f != "__init__.py"
            and not f.startswith("_")
        ]
        return providers

    @staticmethod
    def _get_component_file_path(component_package: str) -> str:
        spec = importlib.util.find_spec(component_package.split(":")[0])
        if spec and spec.origin:
            return os.path.dirname(spec.origin)
        raise NotFoundError(f"Component not found: {component_package}")

    @staticmethod
    def _get_top_level_docstring_from_class(cls):
        file_path = inspect.getfile(cls)
        with open(file_path, "r", encoding="utf-8") as file:
            source_code = file.read()
        parsed_ast = ast.parse(source_code)
        return ast.get_docstring(parsed_ast)

    @staticmethod
    def _get_operations(cls) -> list[OperationSpec]:
        operations: list = []
        for method_name, method in inspect.getmembers(
            cls, predicate=inspect.isfunction
        ):
            if not SpecBuilder._is_operation(method):
                continue

            is_async = inspect.iscoroutinefunction(method)
            if is_async and method_name.startswith("a"):
                name = method_name[1:]
            else:
                name = method_name

            docstring = SpecBuilder._parse_docstring(method.__doc__)
            sig = inspect.signature(method)
            args = []
            parameters: Any = iter(sig.parameters.items())
            if "self" in sig.parameters:
                first_param = next(parameters)
                if first_param[0] == "self":
                    parameters = list(sig.parameters.items())[1:]

            hints = get_type_hints(method)
            for param_name, param in sig.parameters.items():
                if param_name == "self":
                    continue
                if param_name == "kwargs":
                    continue
                param_type = SpecBuilder._get_type_info(
                    hints.get(param_name, None)
                )
                required = param.default == inspect.Parameter.empty
                default = SpecBuilder._get_default_value(param)
                description = None
                if docstring:
                    for p in docstring.params:
                        if p.arg_name == param_name:
                            description = p.description
                            break

                arg_spec = OperationArgSpec(
                    name=param_name,
                    type=param_type,
                    required=required,
                    default=default,
                    description=description,
                )
                args.append(arg_spec)

            ret = None
            if sig.return_annotation != inspect.Signature.empty:
                if docstring and docstring.returns:
                    description = docstring.returns.description
                else:
                    description = None
                ret = OperationReturnSpec(
                    type=SpecBuilder._get_type_info(sig.return_annotation),
                    description=description,
                )
            if docstring:
                description = docstring.short_description
            else:
                description = None

            if not any(op.name == name for op in operations):
                operations.append(
                    OperationSpec(
                        name=name,
                        args=args,
                        ret=ret,
                        description=description,
                    )
                )
        return operations

    @staticmethod
    def _get_type(cls) -> TypeSpec:
        attributes = []
        parsed_doc = SpecBuilder._parse_docstring(inspect.getdoc(cls))
        doc_params = (
            {p.arg_name: p.description for p in parsed_doc.params}
            if parsed_doc
            else {}
        )
        if issubclass(cls, Enum):
            enum_type = type(next(iter(cls)).value)
            for member in cls:
                attributes.append(
                    TypeAttributeSpec(
                        name=member.name,
                        type=SpecBuilder._get_type_info(enum_type),
                        description=doc_params.get(member.name, None),
                        required=False,
                        default=member.value,
                    )
                )
        else:
            type_hints = get_type_hints(cls)
            for name, type_hint in type_hints.items():
                if name.startswith("_"):
                    continue
                if DataModel in cls.__mro__:
                    default = cls.get_default_value(name)
                    if default is Empty:
                        required = True
                        default = None
                    else:
                        required = False
                else:
                    if name in cls.__dict__:
                        default = cls.__dict__[name]
                        required = False
                    else:
                        default = None
                        required = True

                type_ = SpecBuilder._get_type_info(type_hint)
                description = doc_params.get(name, None)
                member_spec = TypeAttributeSpec(
                    name=name,
                    type=type_,
                    description=description,
                    required=required,
                    default=default,
                )
                attributes.append(member_spec)
        description = parsed_doc.description if parsed_doc else None
        inherits = []
        if cls is not DataModel:
            for base in cls.__bases__:
                if base is object:
                    continue
                if base.__module__ in ("builtins", "enum"):
                    continue
                inherits.append(SpecBuilder._get_type(base))
        return TypeSpec(
            name=cls.__name__,
            description=description,
            attributes=attributes,
            inherits=inherits,
        )

    @staticmethod
    def _get_parameters(cls) -> list[ParameterSpec]:
        parameters = []
        init_signature = inspect.signature(cls.__init__)
        doc = inspect.getdoc(cls.__init__)
        doc_parsed = SpecBuilder._parse_docstring(doc) if doc else None
        doc_params = (
            {p.arg_name: p.description for p in doc_parsed.params}
            if doc_parsed
            else {}
        )
        hints = get_type_hints(cls.__init__)
        for param_name, param in init_signature.parameters.items():
            if param_name == "self" or param_name == "kwargs":
                continue
            param_type = SpecBuilder._get_type_info(
                hints.get(param_name, None)
            )
            options = []
            if param.annotation is not inspect.Parameter.empty:
                origin = get_origin(param.annotation)
                args = get_args(param.annotation)
                if origin is Literal:
                    options = [
                        ParameterOptionSpec(label=str(a), value=a)
                        for a in args
                    ]
                elif inspect.isclass(param.annotation) and issubclass(
                    param.annotation, Enum
                ):
                    options = [
                        ParameterOptionSpec(label=e.name, value=e.value)
                        for e in param.annotation
                    ]
            required = param.default == inspect.Parameter.empty
            default = SpecBuilder._get_default_value(param)
            param_spec = ParameterSpec(
                name=param_name,
                type=param_type,
                description=doc_params.get(param_name, None),
                required=required,
                default=default,
                options=options,
            )
            parameters.append(param_spec)
        return parameters

    @staticmethod
    def _parse_docstring(docstring) -> Any:
        try:
            import docstring_parser

            return docstring_parser.parse(docstring)
        except ImportError:
            return None

    @staticmethod
    def _get_type_info(annotation) -> TypeInfo:
        if annotation == inspect.Parameter.empty:
            return TypeInfo(name="Any")
        origin = get_origin(annotation)
        args = get_args(annotation)
        if origin and args:
            if origin.__name__ == "UnionType" or origin.__name__ == "Union":
                return TypeInfo(
                    name="Union",
                    args=[SpecBuilder._get_type_info(arg) for arg in args],
                )
            else:
                return TypeInfo(
                    name=origin.__name__,
                    args=[SpecBuilder._get_type_info(arg) for arg in args],
                )
        if hasattr(annotation, "__name__"):
            if annotation.__name__ == "NoneType":
                return TypeInfo(name="None")
            return TypeInfo(name=annotation.__name__)
        return TypeInfo(name=str(annotation))

    @staticmethod
    def _get_operations_custom_types(cls) -> list[Any]:
        types: list = []
        for _, method in inspect.getmembers(cls, predicate=inspect.isfunction):
            if not SpecBuilder._is_operation(method):
                continue
            custom_types = SpecBuilder._get_args_custom_types(method)
            types = SpecBuilder._combine_types_list(types, custom_types)
        return types

    @staticmethod
    def _get_args_custom_types(method) -> list[Any]:
        types: list = []
        sig = inspect.signature(method)
        hints = get_type_hints(method)
        for param_name, _ in sig.parameters.items():
            custom_types = SpecBuilder._get_custom_types(
                hints.get(param_name, None)
            )
            types = SpecBuilder._combine_types_list(types, custom_types)
        return types

    @staticmethod
    def _get_attributes_custom_types(cls) -> list[Any]:
        types: list = []
        type_hints = get_type_hints(cls)
        for _, type_hint in type_hints.items():
            custom_types = SpecBuilder._get_custom_types(type_hint)
            types = SpecBuilder._combine_types_list(types, custom_types)
        return types

    @staticmethod
    def _get_parameters_custom_types(cls) -> list[Any]:
        types: list = []
        init_signature = inspect.signature(cls.__init__)
        hints = get_type_hints(cls.__init__)
        for param_name, param in init_signature.parameters.items():
            custom_types = SpecBuilder._get_custom_types(
                hints.get(param_name, None)
            )
            types = SpecBuilder._combine_types_list(types, custom_types)
        return types

    @staticmethod
    def _get_provider_parameters_custom_types(
        component_type: str, provider_type: str
    ) -> list[Any]:
        py_path = Loader.get_provider_path(component_type, provider_type)
        try:
            provider = Loader.load_class(py_path, Provider)
            return SpecBuilder._get_parameters_custom_types(provider)
        except Exception:
            return []

    @staticmethod
    def _get_custom_types(annotation) -> list[Any]:
        if annotation == inspect.Parameter.empty:
            return []
        types: list = []
        origin = get_origin(annotation)
        args = get_args(annotation)
        if origin and args:
            for arg in args:
                custom_types = SpecBuilder._get_custom_types(arg)
                types = SpecBuilder._combine_types_list(types, custom_types)
        if hasattr(annotation, "__name__"):
            if (
                annotation.__module__ != "builtins"
                and annotation.__module__ != "typing"
            ):
                types.append(annotation)
        return types

    @staticmethod
    def _combine_types_list(types1: list[Any], types2: list[Any]) -> list[Any]:
        for type in types2:
            if type not in types1:
                types1.append(type)
        return types1

    @staticmethod
    def _get_default_value(param) -> Any:
        default = (
            None if param.default == inspect.Parameter.empty else param.default
        )
        if isinstance(default, Enum):
            return default.value
        return default

    @staticmethod
    def _is_operation(method) -> bool:
        return callable(method) and hasattr(method, "__operation__")

    @staticmethod
    def _add_dependent_types(types: list[Any]) -> list[Any]:
        all_types = list(types)
        while True:
            new_types = []
            for type in all_types:
                attribute_types = SpecBuilder._get_attributes_custom_types(
                    type
                )
                for attr_type in attribute_types:
                    if (
                        attr_type not in all_types
                        and attr_type not in new_types
                    ):
                        new_types.append(attr_type)
            if not new_types:
                break
            all_types.extend(new_types)
        return all_types
